@@ -327,7 +327,8 @@ static inline void shader_setval_inline(gs_shader_param *param,
 
 void gs_shader_set_bool(gs_sparam_t *param, bool val)
 {
-	shader_setval_inline(param, &val, sizeof(bool));
+	int b_val = (int)val;
+	shader_setval_inline(param, &b_val, sizeof(int));
 }
 
 void gs_shader_set_float(gs_sparam_t *param, float val)
@@ -340,7 +341,7 @@ void gs_shader_set_int(gs_sparam_t *param, int val)
 	shader_setval_inline(param, &val, sizeof(int));
 }
 
-void gs_shader_setmatrix3(gs_sparam_t *param, const struct matrix3 *val)
+void gs_shader_set_matrix3(gs_sparam_t *param, const struct matrix3 *val)
 {
 	struct matrix4 mat;
 	matrix4_from_matrix3(&mat, val);
