@@ -18,6 +18,11 @@
 #include "obs.h"
 #include "obs-internal.h"
 
+#define encoder_active(encoder) \
+	os_atomic_load_bool(&encoder->active)
+#define set_encoder_active(encoder, val) \
+	os_atomic_set_bool(&encoder->active, val)
+
 struct obs_encoder_info *find_encoder(const char *id)
 {
 	for (size_t i = 0; i < obs->encoder_types.num; i++) {
@@ -33,18 +38,28 @@ struct obs_encoder_info *find_encoder(const char *id)
 const char *obs_encoder_get_display_name(const char *id)
 {
 	struct obs_encoder_info *ei = find_encoder(id);
-	return ei ? ei->get_name() : NULL;
+	return ei ? ei->get_name(ei->type_data) : NULL;
 }
 
 static bool init_encoder(struct obs_encoder *encoder, const char *name,
-		obs_data_t *settings)
+		obs_data_t *settings, obs_data_t *hotkey_data)
 {
+	pthread_mutexattr_t attr;
+
+	pthread_mutex_init_value(&encoder->init_mutex);
 	pthread_mutex_init_value(&encoder->callbacks_mutex);
 	pthread_mutex_init_value(&encoder->outputs_mutex);
 
-	if (!obs_context_data_init(&encoder->context, settings, name))
+	if (pthread_mutexattr_init(&attr) != 0)
 		return false;
-	if (pthread_mutex_init(&encoder->callbacks_mutex, NULL) != 0)
+	if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE) != 0)
+		return false;
+	if (!obs_context_data_init(&encoder->context, OBS_OBJ_TYPE_ENCODER,
+				settings, name, hotkey_data, false))
+		return false;
+	if (pthread_mutex_init(&encoder->init_mutex, &attr) != 0)
+		return false;
+	if (pthread_mutex_init(&encoder->callbacks_mutex, &attr) != 0)
 		return false;
 	if (pthread_mutex_init(&encoder->outputs_mutex, NULL) != 0)
 		return false;
@@ -57,25 +72,38 @@ static bool init_encoder(struct obs_encoder *encoder, const char *name,
 
 static struct obs_encoder *create_encoder(const char *id,
 		enum obs_encoder_type type, const char *name,
-		obs_data_t *settings, size_t mixer_idx)
+		obs_data_t *settings, size_t mixer_idx, obs_data_t *hotkey_data)
 {
 	struct obs_encoder *encoder;
 	struct obs_encoder_info *ei = find_encoder(id);
 	bool success;
 
-	if (!ei || ei->type != type)
+	if (ei && ei->type != type)
 		return NULL;
 
 	encoder = bzalloc(sizeof(struct obs_encoder));
-	encoder->info = *ei;
 	encoder->mixer_idx = mixer_idx;
 
-	success = init_encoder(encoder, name, settings);
-	if (!success) {
-		obs_encoder_destroy(encoder);
-		encoder = NULL;
+	if (!ei) {
+		blog(LOG_ERROR, "Encoder ID '%s' not found", id);
+
+		encoder->info.id      = bstrdup(id);
+		encoder->info.type    = type;
+		encoder->owns_info_id = true;
+	} else {
+		encoder->info = *ei;
 	}
 
+	success = init_encoder(encoder, name, settings, hotkey_data);
+	if (!success) {
+		blog(LOG_ERROR, "creating encoder '%s' (%s) failed", name, id);
+		obs_encoder_destroy(encoder);
+		return NULL;
+	}
+
+	encoder->control = bzalloc(sizeof(obs_weak_encoder_t));
+	encoder->control->encoder = encoder;
+
 	obs_context_data_insert(&encoder->context,
 			&obs->data.encoders_mutex,
 			&obs->data.first_encoder);
@@ -85,32 +113,29 @@ static struct obs_encoder *create_encoder(const char *id,
 }
 
 obs_encoder_t *obs_video_encoder_create(const char *id, const char *name,
-		obs_data_t *settings)
+		obs_data_t *settings, obs_data_t *hotkey_data)
 {
 	if (!name || !id) return NULL;
-	return create_encoder(id, OBS_ENCODER_VIDEO, name, settings, 0);
+	return create_encoder(id, OBS_ENCODER_VIDEO, name, settings, 0,
+			hotkey_data);
 }
 
 obs_encoder_t *obs_audio_encoder_create(const char *id, const char *name,
-		obs_data_t *settings, size_t mixer_idx)
+		obs_data_t *settings, size_t mixer_idx, obs_data_t *hotkey_data)
 {
 	if (!name || !id) return NULL;
-	return create_encoder(id, OBS_ENCODER_AUDIO, name, settings, mixer_idx);
+	return create_encoder(id, OBS_ENCODER_AUDIO, name, settings, mixer_idx,
+			hotkey_data);
 }
 
 static void receive_video(void *param, struct video_data *frame);
 static void receive_audio(void *param, size_t mix_idx, struct audio_data *data);
 
-static inline struct audio_convert_info *get_audio_info(
-		const struct obs_encoder *encoder,
+static inline void get_audio_info(const struct obs_encoder *encoder,
 		struct audio_convert_info *info)
 {
 	const struct audio_output_info *aoi;
 	aoi = audio_output_get_info(encoder->media);
-	memset(info, 0, sizeof(struct audio_convert_info));
-
-	if (encoder->info.get_audio_info)
-		encoder->info.get_audio_info(encoder->context.data, info);
 
 	if (info->format == AUDIO_FORMAT_UNKNOWN)
 		info->format = aoi->format;
@@ -119,18 +144,27 @@ static inline struct audio_convert_info *get_audio_info(
 	if (info->speakers == SPEAKERS_UNKNOWN)
 		info->speakers = aoi->speakers;
 
-	return info;
+	if (encoder->info.get_audio_info)
+		encoder->info.get_audio_info(encoder->context.data, info);
 }
 
-static inline struct video_scale_info *get_video_info(
-		const struct obs_encoder *encoder,
+static inline void get_video_info(struct obs_encoder *encoder,
 		struct video_scale_info *info)
 {
-	if (encoder->info.get_video_info)
-		if (encoder->info.get_video_info(encoder->context.data, info))
-			return info;
+	const struct video_output_info *voi;
+	voi = video_output_get_info(encoder->media);
 
-	return NULL;
+	info->format     = voi->format;
+	info->colorspace = voi->colorspace;
+	info->range      = voi->range;
+	info->width      = obs_encoder_get_width(encoder);
+	info->height     = obs_encoder_get_height(encoder);
+
+	if (encoder->info.get_video_info)
+		encoder->info.get_video_info(encoder->context.data, info);
+
+	if (info->width != voi->width || info->height != voi->height)
+		obs_encoder_set_scaled_size(encoder, info->width, info->height);
 }
 
 static inline bool has_scaling(const struct obs_encoder *encoder)
@@ -145,37 +179,21 @@ static inline bool has_scaling(const struct obs_encoder *encoder)
 
 static void add_connection(struct obs_encoder *encoder)
 {
-	struct audio_convert_info audio_info = {0};
-	struct video_scale_info   video_info = {0};
-
 	if (encoder->info.type == OBS_ENCODER_AUDIO) {
+		struct audio_convert_info audio_info = {0};
 		get_audio_info(encoder, &audio_info);
+
 		audio_output_connect(encoder->media, encoder->mixer_idx,
 				&audio_info, receive_audio, encoder);
 	} else {
-		struct video_scale_info *info =
-			get_video_info(encoder, &video_info);
+		struct video_scale_info info = {0};
+		get_video_info(encoder, &info);
 
-		if (!info && has_scaling(encoder)) {
-			const struct video_output_info *output_info;
-			output_info = video_output_get_info(encoder->media);
-
-			info             = &video_info;
-			info->format     = output_info->format;
-			info->colorspace = output_info->colorspace;
-			info->range      = output_info->range;
-		}
-
-		if (info && (!info->width || !info->height)) {
-			info->width  = obs_encoder_get_width(encoder);
-			info->height = obs_encoder_get_height(encoder);
-		}
-
-		video_output_connect(encoder->media, info, receive_video,
+		video_output_connect(encoder->media, &info, receive_video,
 			encoder);
 	}
 
-	encoder->active = true;
+	set_encoder_active(encoder, true);
 }
 
 static void remove_connection(struct obs_encoder *encoder)
@@ -187,7 +205,8 @@ static void remove_connection(struct obs_encoder *encoder)
 		video_output_disconnect(encoder->media, receive_video,
 				encoder);
 
-	encoder->active = false;
+	obs_encoder_shutdown(encoder);
+	set_encoder_active(encoder, false);
 }
 
 static inline void free_audio_buffers(struct obs_encoder *encoder)
@@ -217,9 +236,12 @@ static void obs_encoder_actually_destroy(obs_encoder_t *encoder)
 		if (encoder->context.data)
 			encoder->info.destroy(encoder->context.data);
 		da_free(encoder->callbacks);
+		pthread_mutex_destroy(&encoder->init_mutex);
 		pthread_mutex_destroy(&encoder->callbacks_mutex);
 		pthread_mutex_destroy(&encoder->outputs_mutex);
 		obs_context_data_free(&encoder->context);
+		if (encoder->owns_info_id)
+			bfree((void*)encoder->info.id);
 		bfree(encoder);
 	}
 }
@@ -246,12 +268,14 @@ void obs_encoder_destroy(obs_encoder_t *encoder)
 
 const char *obs_encoder_get_name(const obs_encoder_t *encoder)
 {
-	return encoder ? encoder->context.name : NULL;
+	return obs_encoder_valid(encoder, "obs_encoder_get_name") ?
+		encoder->context.name : NULL;
 }
 
 void obs_encoder_set_name(obs_encoder_t *encoder, const char *name)
 {
-	if (!encoder) return;
+	if (!obs_encoder_valid(encoder, "obs_encoder_set_name"))
+		return;
 
 	if (name && *name && strcmp(name, encoder->context.name) != 0)
 		obs_context_data_setname(&encoder->context, name);
@@ -288,7 +312,10 @@ obs_properties_t *obs_get_encoder_properties(const char *id)
 
 obs_properties_t *obs_encoder_properties(const obs_encoder_t *encoder)
 {
-	if (encoder && encoder->info.get_properties) {
+	if (!obs_encoder_valid(encoder, "obs_encoder_properties"))
+		return NULL;
+
+	if (encoder->info.get_properties) {
 		obs_properties_t *props;
 		props = encoder->info.get_properties(encoder->context.data);
 		obs_properties_apply_settings(props, encoder->context.settings);
@@ -299,7 +326,8 @@ obs_properties_t *obs_encoder_properties(const obs_encoder_t *encoder)
 
 void obs_encoder_update(obs_encoder_t *encoder, obs_data_t *settings)
 {
-	if (!encoder) return;
+	if (!obs_encoder_valid(encoder, "obs_encoder_update"))
+		return;
 
 	obs_data_apply(encoder->context.settings, settings);
 
@@ -311,7 +339,10 @@ void obs_encoder_update(obs_encoder_t *encoder, obs_data_t *settings)
 bool obs_encoder_get_extra_data(const obs_encoder_t *encoder,
 		uint8_t **extra_data, size_t *size)
 {
-	if (encoder && encoder->info.get_extra_data && encoder->context.data)
+	if (!obs_encoder_valid(encoder, "obs_encoder_get_extra_data"))
+		return false;
+
+	if (encoder->info.get_extra_data && encoder->context.data)
 		return encoder->info.get_extra_data(encoder->context.data,
 				extra_data, size);
 
@@ -320,7 +351,8 @@ bool obs_encoder_get_extra_data(const obs_encoder_t *encoder,
 
 obs_data_t *obs_encoder_get_settings(const obs_encoder_t *encoder)
 {
-	if (!encoder) return NULL;
+	if (!obs_encoder_valid(encoder, "obs_encoder_get_settings"))
+		return NULL;
 
 	obs_data_addref(encoder->context.settings);
 	return encoder->context.settings;
@@ -337,7 +369,7 @@ static inline void reset_audio_buffers(struct obs_encoder *encoder)
 
 static void intitialize_audio_encoder(struct obs_encoder *encoder)
 {
-	struct audio_convert_info info;
+	struct audio_convert_info info = {0};
 	get_audio_info(encoder, &info);
 
 	encoder->samplerate = info.samples_per_sec;
@@ -350,30 +382,55 @@ static void intitialize_audio_encoder(struct obs_encoder *encoder)
 	reset_audio_buffers(encoder);
 }
 
-bool obs_encoder_initialize(obs_encoder_t *encoder)
+static inline bool obs_encoder_initialize_internal(obs_encoder_t *encoder)
 {
-	if (!encoder) return false;
-
-	if (encoder->active)
+	if (encoder_active(encoder))
+		return true;
+	if (encoder->initialized)
 		return true;
 
-	if (encoder->context.data)
-		encoder->info.destroy(encoder->context.data);
+	obs_encoder_shutdown(encoder);
 
-	encoder->context.data = encoder->info.create(encoder->context.settings,
-			encoder);
+	if (encoder->info.create)
+		encoder->context.data = encoder->info.create(
+				encoder->context.settings, encoder);
 	if (!encoder->context.data)
 		return false;
 
-	encoder->paired_encoder  = NULL;
-	encoder->start_ts        = 0;
-
 	if (encoder->info.type == OBS_ENCODER_AUDIO)
 		intitialize_audio_encoder(encoder);
 
+	encoder->initialized = true;
 	return true;
 }
 
+bool obs_encoder_initialize(obs_encoder_t *encoder)
+{
+	bool success;
+
+	if (!encoder) return false;
+
+	pthread_mutex_lock(&encoder->init_mutex);
+	success = obs_encoder_initialize_internal(encoder);
+	pthread_mutex_unlock(&encoder->init_mutex);
+
+	return success;
+}
+
+void obs_encoder_shutdown(obs_encoder_t *encoder)
+{
+	pthread_mutex_lock(&encoder->init_mutex);
+	if (encoder->context.data) {
+		encoder->info.destroy(encoder->context.data);
+		encoder->context.data    = NULL;
+		encoder->paired_encoder  = NULL;
+		encoder->first_received  = false;
+		encoder->offset_usec     = 0;
+		encoder->start_ts        = 0;
+	}
+	pthread_mutex_unlock(&encoder->init_mutex);
+}
+
 static inline size_t get_callback_idx(
 		const struct obs_encoder *encoder,
 		void (*new_packet)(void *param, struct encoder_packet *packet),
@@ -389,14 +446,15 @@ static inline size_t get_callback_idx(
 	return DARRAY_INVALID;
 }
 
-void obs_encoder_start(obs_encoder_t *encoder,
+static inline void obs_encoder_start_internal(obs_encoder_t *encoder,
 		void (*new_packet)(void *param, struct encoder_packet *packet),
 		void *param)
 {
 	struct encoder_callback cb = {false, new_packet, param};
 	bool first   = false;
 
-	if (!encoder || !new_packet || !encoder->context.data) return;
+	if (!encoder->context.data)
+		return;
 
 	pthread_mutex_lock(&encoder->callbacks_mutex);
 
@@ -414,15 +472,27 @@ void obs_encoder_start(obs_encoder_t *encoder,
 	}
 }
 
-void obs_encoder_stop(obs_encoder_t *encoder,
+void obs_encoder_start(obs_encoder_t *encoder,
+		void (*new_packet)(void *param, struct encoder_packet *packet),
+		void *param)
+{
+	if (!obs_encoder_valid(encoder, "obs_encoder_start"))
+		return;
+	if (!obs_ptr_valid(new_packet, "obs_encoder_start"))
+		return;
+
+	pthread_mutex_lock(&encoder->init_mutex);
+	obs_encoder_start_internal(encoder, new_packet, param);
+	pthread_mutex_unlock(&encoder->init_mutex);
+}
+
+static inline bool obs_encoder_stop_internal(obs_encoder_t *encoder,
 		void (*new_packet)(void *param, struct encoder_packet *packet),
 		void *param)
 {
 	bool   last = false;
 	size_t idx;
 
-	if (!encoder) return;
-
 	pthread_mutex_lock(&encoder->callbacks_mutex);
 
 	idx = get_callback_idx(encoder, new_packet, param);
@@ -435,15 +505,39 @@ void obs_encoder_stop(obs_encoder_t *encoder,
 
 	if (last) {
 		remove_connection(encoder);
+		encoder->initialized = false;
 
-		if (encoder->destroy_on_stop)
+		if (encoder->destroy_on_stop) {
+			pthread_mutex_unlock(&encoder->init_mutex);
 			obs_encoder_actually_destroy(encoder);
+			return true;
+		}
 	}
+
+	return false;
+}
+
+void obs_encoder_stop(obs_encoder_t *encoder,
+		void (*new_packet)(void *param, struct encoder_packet *packet),
+		void *param)
+{
+	bool destroyed;
+
+	if (!obs_encoder_valid(encoder, "obs_encoder_stop"))
+		return;
+	if (!obs_ptr_valid(new_packet, "obs_encoder_stop"))
+		return;
+
+	pthread_mutex_lock(&encoder->init_mutex);
+	destroyed = obs_encoder_stop_internal(encoder, new_packet, param);
+	if (!destroyed)
+		pthread_mutex_unlock(&encoder->init_mutex);
 }
 
 const char *obs_encoder_get_codec(const obs_encoder_t *encoder)
 {
-	return encoder ? encoder->info.codec : NULL;
+	return obs_encoder_valid(encoder, "obs_encoder_get_codec") ?
+		encoder->info.codec : NULL;
 }
 
 const char *obs_get_encoder_codec(const char *id)
@@ -454,7 +548,8 @@ const char *obs_get_encoder_codec(const char *id)
 
 enum obs_encoder_type obs_encoder_get_type(const obs_encoder_t *encoder)
 {
-	return encoder ? encoder->info.type : OBS_ENCODER_AUDIO;
+	return obs_encoder_valid(encoder, "obs_encoder_get_type") ?
+		encoder->info.type : OBS_ENCODER_AUDIO;
 }
 
 enum obs_encoder_type obs_get_encoder_type(const char *id)
@@ -466,10 +561,15 @@ enum obs_encoder_type obs_get_encoder_type(const char *id)
 void obs_encoder_set_scaled_size(obs_encoder_t *encoder, uint32_t width,
 		uint32_t height)
 {
-	if (!encoder || encoder->info.type != OBS_ENCODER_VIDEO)
+	if (!obs_encoder_valid(encoder, "obs_encoder_set_scaled_size"))
 		return;
-
-	if (encoder->active) {
+	if (encoder->info.type != OBS_ENCODER_VIDEO) {
+		blog(LOG_WARNING, "obs_encoder_set_scaled_size: "
+				"encoder '%s' is not a video encoder",
+				obs_encoder_get_name(encoder));
+		return;
+	}
+	if (encoder_active(encoder)) {
 		blog(LOG_WARNING, "encoder '%s': Cannot set the scaled "
 		                  "resolution while the encoder is active",
 		                  obs_encoder_get_name(encoder));
@@ -482,8 +582,15 @@ void obs_encoder_set_scaled_size(obs_encoder_t *encoder, uint32_t width,
 
 uint32_t obs_encoder_get_width(const obs_encoder_t *encoder)
 {
-	if (!encoder || !encoder->media ||
-	    encoder->info.type != OBS_ENCODER_VIDEO)
+	if (!obs_encoder_valid(encoder, "obs_encoder_get_width"))
+		return 0;
+	if (encoder->info.type != OBS_ENCODER_VIDEO) {
+		blog(LOG_WARNING, "obs_encoder_get_width: "
+				"encoder '%s' is not a video encoder",
+				obs_encoder_get_name(encoder));
+		return 0;
+	}
+	if (!encoder->media)
 		return 0;
 
 	return encoder->scaled_width != 0 ?
@@ -493,8 +600,15 @@ uint32_t obs_encoder_get_width(const obs_encoder_t *encoder)
 
 uint32_t obs_encoder_get_height(const obs_encoder_t *encoder)
 {
-	if (!encoder || !encoder->media ||
-	    encoder->info.type != OBS_ENCODER_VIDEO)
+	if (!obs_encoder_valid(encoder, "obs_encoder_get_height"))
+		return 0;
+	if (encoder->info.type != OBS_ENCODER_VIDEO) {
+		blog(LOG_WARNING, "obs_encoder_get_height: "
+				"encoder '%s' is not a video encoder",
+				obs_encoder_get_name(encoder));
+		return 0;
+	}
+	if (!encoder->media)
 		return 0;
 
 	return encoder->scaled_width != 0 ?
@@ -502,11 +616,37 @@ uint32_t obs_encoder_get_height(const obs_encoder_t *encoder)
 		video_output_get_height(encoder->media);
 }
 
+uint32_t obs_encoder_get_sample_rate(const obs_encoder_t *encoder)
+{
+	if (!obs_encoder_valid(encoder, "obs_encoder_get_sample_rate"))
+		return 0;
+	if (encoder->info.type != OBS_ENCODER_AUDIO) {
+		blog(LOG_WARNING, "obs_encoder_get_sample_rate: "
+				"encoder '%s' is not an audio encoder",
+				obs_encoder_get_name(encoder));
+		return 0;
+	}
+	if (!encoder->media)
+		return 0;
+
+	return encoder->samplerate != 0 ?
+		encoder->samplerate :
+		audio_output_get_sample_rate(encoder->media);
+}
+
 void obs_encoder_set_video(obs_encoder_t *encoder, video_t *video)
 {
 	const struct video_output_info *voi;
 
-	if (!video || !encoder || encoder->info.type != OBS_ENCODER_VIDEO)
+	if (!obs_encoder_valid(encoder, "obs_encoder_set_video"))
+		return;
+	if (encoder->info.type != OBS_ENCODER_VIDEO) {
+		blog(LOG_WARNING, "obs_encoder_set_video: "
+				"encoder '%s' is not a video encoder",
+				obs_encoder_get_name(encoder));
+		return;
+	}
+	if (!video)
 		return;
 
 	voi = video_output_get_info(video);
@@ -518,7 +658,15 @@ void obs_encoder_set_video(obs_encoder_t *encoder, video_t *video)
 
 void obs_encoder_set_audio(obs_encoder_t *encoder, audio_t *audio)
 {
-	if (!audio || !encoder || encoder->info.type != OBS_ENCODER_AUDIO)
+	if (!obs_encoder_valid(encoder, "obs_encoder_set_audio"))
+		return;
+	if (encoder->info.type != OBS_ENCODER_AUDIO) {
+		blog(LOG_WARNING, "obs_encoder_set_audio: "
+				"encoder '%s' is not an audio encoder",
+				obs_encoder_get_name(encoder));
+		return;
+	}
+	if (!audio)
 		return;
 
 	encoder->media        = audio;
@@ -528,19 +676,36 @@ void obs_encoder_set_audio(obs_encoder_t *encoder, audio_t *audio)
 
 video_t *obs_encoder_video(const obs_encoder_t *encoder)
 {
-	return (encoder && encoder->info.type == OBS_ENCODER_VIDEO) ?
-		encoder->media : NULL;
+	if (!obs_encoder_valid(encoder, "obs_encoder_video"))
+		return NULL;
+	if (encoder->info.type != OBS_ENCODER_VIDEO) {
+		blog(LOG_WARNING, "obs_encoder_set_video: "
+				"encoder '%s' is not a video encoder",
+				obs_encoder_get_name(encoder));
+		return NULL;
+	}
+
+	return encoder->media;
 }
 
 audio_t *obs_encoder_audio(const obs_encoder_t *encoder)
 {
-	return (encoder && encoder->info.type == OBS_ENCODER_AUDIO) ?
-		encoder->media : NULL;
+	if (!obs_encoder_valid(encoder, "obs_encoder_audio"))
+		return NULL;
+	if (encoder->info.type != OBS_ENCODER_AUDIO) {
+		blog(LOG_WARNING, "obs_encoder_set_audio: "
+				"encoder '%s' is not an audio encoder",
+				obs_encoder_get_name(encoder));
+		return NULL;
+	}
+
+	return encoder->media;
 }
 
 bool obs_encoder_active(const obs_encoder_t *encoder)
 {
-	return encoder ? encoder->active : false;
+	return obs_encoder_valid(encoder, "obs_encoder_active") ?
+		encoder_active(encoder) : false;
 }
 
 static inline bool get_sei(const struct obs_encoder *encoder,
@@ -566,7 +731,7 @@ static void send_first_video_packet(struct obs_encoder *encoder,
 
 	da_init(data);
 
-	if (!get_sei(encoder, &sei, &size)) {
+	if (!get_sei(encoder, &sei, &size) || !sei || !size) {
 		cb->new_packet(cb->param, packet);
 		cb->sent_first_packet = true;
 		return;
@@ -605,9 +770,16 @@ static void full_stop(struct obs_encoder *encoder)
 	}
 }
 
+static const char *do_encode_name = "do_encode";
 static inline void do_encode(struct obs_encoder *encoder,
 		struct encoder_frame *frame)
 {
+	profile_start(do_encode_name);
+	if (!encoder->profile_encoder_encode_name)
+		encoder->profile_encoder_encode_name =
+			profile_store_name(obs_get_profiler_name_store(),
+					"encode(%s)", encoder->context.name);
+
 	struct encoder_packet pkt = {0};
 	bool received = false;
 	bool success;
@@ -616,8 +788,10 @@ static inline void do_encode(struct obs_encoder *encoder,
 	pkt.timebase_den = encoder->timebase_den;
 	pkt.encoder = encoder;
 
+	profile_start(encoder->profile_encoder_encode_name);
 	success = encoder->info.encode(encoder->context.data, frame, &pkt,
 			&received);
+	profile_end(encoder->profile_encoder_encode_name);
 	if (!success) {
 		full_stop(encoder);
 		blog(LOG_ERROR, "Error encoding with encoder '%s'",
@@ -626,27 +800,46 @@ static inline void do_encode(struct obs_encoder *encoder,
 	}
 
 	if (received) {
+		if (!encoder->first_received) {
+			encoder->offset_usec = packet_dts_usec(&pkt);
+			encoder->first_received = true;
+		}
+
 		/* we use system time here to ensure sync with other encoders,
 		 * you do not want to use relative timestamps here */
-		pkt.dts_usec = encoder->start_ts / 1000 + packet_dts_usec(&pkt);
+		pkt.dts_usec = encoder->start_ts / 1000 +
+			packet_dts_usec(&pkt) - encoder->offset_usec;
 
 		pthread_mutex_lock(&encoder->callbacks_mutex);
 
-		for (size_t i = 0; i < encoder->callbacks.num; i++) {
+		for (size_t i = encoder->callbacks.num; i > 0; i--) {
 			struct encoder_callback *cb;
-			cb = encoder->callbacks.array+i;
+			cb = encoder->callbacks.array+(i-1);
 			send_packet(encoder, cb, &pkt);
 		}
 
 		pthread_mutex_unlock(&encoder->callbacks_mutex);
 	}
+
+	profile_end(do_encode_name);
 }
 
+static const char *receive_video_name = "receive_video";
 static void receive_video(void *param, struct video_data *frame)
 {
+	profile_start(receive_video_name);
+
 	struct obs_encoder    *encoder  = param;
+	struct obs_encoder    *pair     = encoder->paired_encoder;
 	struct encoder_frame  enc_frame;
 
+	if (!encoder->first_received && pair) {
+		if (!pair->first_received ||
+		    pair->first_raw_ts > frame->timestamp) {
+			goto wait_for_audio;
+		}
+	}
+
 	memset(&enc_frame, 0, sizeof(struct encoder_frame));
 
 	for (size_t i = 0; i < MAX_AV_PLANES; i++) {
@@ -663,41 +856,20 @@ static void receive_video(void *param, struct video_data *frame)
 	do_encode(encoder, &enc_frame);
 
 	encoder->cur_pts += encoder->timebase_num;
+
+wait_for_audio:
+	profile_end(receive_video_name);
 }
 
-static bool buffer_audio(struct obs_encoder *encoder, struct audio_data *data)
+static void clear_audio(struct obs_encoder *encoder)
 {
-	size_t samplerate = encoder->samplerate;
-	size_t size = data->frames * encoder->blocksize;
-	size_t offset_size = 0;
-
-	if (!encoder->start_ts && encoder->paired_encoder) {
-		uint64_t end_ts     = data->timestamp;
-		uint64_t v_start_ts = encoder->paired_encoder->start_ts;
-
-		/* no video yet, so don't start audio */
-		if (!v_start_ts)
-			return false;
-
-		/* audio starting point still not synced with video starting
-		 * point, so don't start audio */
-		end_ts += (uint64_t)data->frames * 1000000000ULL / samplerate;
-		if (end_ts <= v_start_ts)
-			return false;
-
-		/* ready to start audio, truncate if necessary */
-		if (data->timestamp < v_start_ts) {
-			uint64_t offset = v_start_ts - data->timestamp;
-			offset = (int)(offset * samplerate / 1000000000);
-			offset_size = (size_t)offset * encoder->blocksize;
-		}
-
-		encoder->start_ts = v_start_ts;
-
-	} else if (!encoder->start_ts && !encoder->paired_encoder) {
-		encoder->start_ts = data->timestamp;
-	}
+	for (size_t i = 0; i < encoder->planes; i++)
+		circlebuf_free(&encoder->audio_input_buffer[i]);
+}
 
+static inline void push_back_audio(struct obs_encoder *encoder,
+		struct audio_data *data, size_t size, size_t offset_size)
+{
 	size -= offset_size;
 
 	/* push in to the circular buffer */
@@ -705,8 +877,92 @@ static bool buffer_audio(struct obs_encoder *encoder, struct audio_data *data)
 		for (size_t i = 0; i < encoder->planes; i++)
 			circlebuf_push_back(&encoder->audio_input_buffer[i],
 					data->data[i] + offset_size, size);
+}
 
-	return true;
+static inline size_t calc_offset_size(struct obs_encoder *encoder,
+		uint64_t v_start_ts, uint64_t a_start_ts)
+{
+	uint64_t offset = v_start_ts - a_start_ts;
+	offset = (uint64_t)offset * (uint64_t)encoder->samplerate /
+		1000000000ULL;
+	return (size_t)offset * encoder->blocksize;
+}
+
+static void start_from_buffer(struct obs_encoder *encoder, uint64_t v_start_ts)
+{
+	size_t size = encoder->audio_input_buffer[0].size;
+	struct audio_data audio = {0};
+	size_t offset_size = 0;
+
+	for (size_t i = 0; i < MAX_AV_PLANES; i++) {
+		audio.data[i] = encoder->audio_input_buffer[i].data;
+		memset(&encoder->audio_input_buffer[i], 0,
+				sizeof(struct circlebuf));
+	}
+
+	if (encoder->first_raw_ts < v_start_ts)
+		offset_size = calc_offset_size(encoder, v_start_ts,
+				encoder->first_raw_ts);
+
+	push_back_audio(encoder, &audio, size, offset_size);
+
+	for (size_t i = 0; i < MAX_AV_PLANES; i++)
+		bfree(audio.data[i]);
+}
+
+static const char *buffer_audio_name = "buffer_audio";
+static bool buffer_audio(struct obs_encoder *encoder, struct audio_data *data)
+{
+	profile_start(buffer_audio_name);
+
+	size_t size = data->frames * encoder->blocksize;
+	size_t offset_size = 0;
+	bool success = true;
+
+	if (!encoder->start_ts && encoder->paired_encoder) {
+		uint64_t end_ts     = data->timestamp;
+		uint64_t v_start_ts = encoder->paired_encoder->start_ts;
+
+		/* no video yet, so don't start audio */
+		if (!v_start_ts) {
+			success = false;
+			goto fail;
+		}
+
+		/* audio starting point still not synced with video starting
+		 * point, so don't start audio */
+		end_ts += (uint64_t)data->frames * 1000000000ULL /
+			(uint64_t)encoder->samplerate;
+		if (end_ts <= v_start_ts) {
+			success = false;
+			goto fail;
+		}
+
+		/* ready to start audio, truncate if necessary */
+		if (data->timestamp < v_start_ts)
+			offset_size = calc_offset_size(encoder, v_start_ts,
+					data->timestamp);
+		if (data->timestamp <= v_start_ts)
+			clear_audio(encoder);
+
+		encoder->start_ts = v_start_ts;
+
+		/* use currently buffered audio instead */
+		if (v_start_ts < data->timestamp) {
+			start_from_buffer(encoder, v_start_ts);
+			goto skip_push;
+		}
+
+	} else if (!encoder->start_ts && !encoder->paired_encoder) {
+		encoder->start_ts = data->timestamp;
+	}
+
+fail:
+	push_back_audio(encoder, data, size, offset_size);
+
+skip_push:
+	profile_end(buffer_audio_name);
+	return success;
 }
 
 static void send_audio_data(struct obs_encoder *encoder)
@@ -732,17 +988,29 @@ static void send_audio_data(struct obs_encoder *encoder)
 	encoder->cur_pts += encoder->framesize;
 }
 
+static const char *receive_audio_name = "receive_audio";
 static void receive_audio(void *param, size_t mix_idx, struct audio_data *data)
 {
+	profile_start(receive_audio_name);
+
 	struct obs_encoder *encoder = param;
 
+	if (!encoder->first_received) {
+		encoder->first_raw_ts = data->timestamp;
+		encoder->first_received = true;
+		clear_audio(encoder);
+	}
+
 	if (!buffer_audio(encoder, data))
-		return;
+		goto end;
 
 	while (encoder->audio_input_buffer[0].size >= encoder->framesize_bytes)
 		send_audio_data(encoder);
 
 	UNUSED_PARAMETER(mix_idx);
+
+end:
+	profile_end(receive_audio_name);
 }
 
 void obs_encoder_add_output(struct obs_encoder *encoder,
@@ -777,3 +1045,114 @@ void obs_free_encoder_packet(struct encoder_packet *packet)
 	bfree(packet->data);
 	memset(packet, 0, sizeof(struct encoder_packet));
 }
+
+void obs_encoder_set_preferred_video_format(obs_encoder_t *encoder,
+		enum video_format format)
+{
+	if (!encoder || encoder->info.type != OBS_ENCODER_VIDEO)
+		return;
+
+	encoder->preferred_format = format;
+}
+
+enum video_format obs_encoder_get_preferred_video_format(
+		const obs_encoder_t *encoder)
+{
+	if (!encoder || encoder->info.type != OBS_ENCODER_VIDEO)
+		return VIDEO_FORMAT_NONE;
+
+	return encoder->preferred_format;
+}
+
+void obs_encoder_addref(obs_encoder_t *encoder)
+{
+	if (!encoder)
+		return;
+
+	obs_ref_addref(&encoder->control->ref);
+}
+
+void obs_encoder_release(obs_encoder_t *encoder)
+{
+	if (!encoder)
+		return;
+
+	obs_weak_encoder_t *control = encoder->control;
+	if (obs_ref_release(&control->ref)) {
+		// The order of operations is important here since
+		// get_context_by_name in obs.c relies on weak refs
+		// being alive while the context is listed
+		obs_encoder_destroy(encoder);
+		obs_weak_encoder_release(control);
+	}
+}
+
+void obs_weak_encoder_addref(obs_weak_encoder_t *weak)
+{
+	if (!weak)
+		return;
+
+	obs_weak_ref_addref(&weak->ref);
+}
+
+void obs_weak_encoder_release(obs_weak_encoder_t *weak)
+{
+	if (!weak)
+		return;
+
+	if (obs_weak_ref_release(&weak->ref))
+		bfree(weak);
+}
+
+obs_encoder_t *obs_encoder_get_ref(obs_encoder_t *encoder)
+{
+	if (!encoder)
+		return NULL;
+
+	return obs_weak_encoder_get_encoder(encoder->control);
+}
+
+obs_weak_encoder_t *obs_encoder_get_weak_encoder(obs_encoder_t *encoder)
+{
+	if (!encoder)
+		return NULL;
+
+	obs_weak_encoder_t *weak = encoder->control;
+	obs_weak_encoder_addref(weak);
+	return weak;
+}
+
+obs_encoder_t *obs_weak_encoder_get_encoder(obs_weak_encoder_t *weak)
+{
+	if (!weak)
+		return NULL;
+
+	if (obs_weak_ref_get_ref(&weak->ref))
+		return weak->encoder;
+
+	return NULL;
+}
+
+bool obs_weak_encoder_references_encoder(obs_weak_encoder_t *weak,
+		obs_encoder_t *encoder)
+{
+	return weak && encoder && weak->encoder == encoder;
+}
+
+void *obs_encoder_get_type_data(obs_encoder_t *encoder)
+{
+	return obs_encoder_valid(encoder, "obs_encoder_get_type_data")
+		? encoder->info.type_data : NULL;
+}
+
+const char *obs_encoder_get_id(const obs_encoder_t *encoder)
+{
+	return obs_encoder_valid(encoder, "obs_encoder_get_id")
+		? encoder->info.id : NULL;
+}
+
+uint32_t obs_get_encoder_caps(const char *encoder_id)
+{
+	struct obs_encoder_info *info = find_encoder(encoder_id);
+	return info ? info->caps : 0;
+}
