@@ -21,39 +21,9 @@
 #include "media-io/format-conversion.h"
 #include "media-io/video-frame.h"
 
-static inline void calculate_base_volume(struct obs_core_data *data,
-		struct obs_view *view, obs_source_t *target)
-{
-	if (!target->activate_refs) {
-		target->base_volume = 0.0f;
-
-	/* only walk the tree if there are transitions active */
-	} else if (data->active_transitions) {
-		float best_vol = 0.0f;
-
-		for (size_t i = 0; i < MAX_CHANNELS; i++) {
-			struct obs_source *source = view->channels[i];
-			float vol = 0.0f;
-
-			if (!source)
-				continue;
-
-			vol = obs_source_get_target_volume(source, target);
-			if (best_vol < vol)
-				best_vol = vol;
-		}
-
-		target->base_volume = best_vol;
-
-	} else {
-		target->base_volume = 1.0f;
-	}
-}
-
 static uint64_t tick_sources(uint64_t cur_time, uint64_t last_time)
 {
 	struct obs_core_data *data = &obs->data;
-	struct obs_view      *view = &data->main_view;
 	struct obs_source    *source;
 	uint64_t             delta_time;
 	float                seconds;
@@ -70,23 +40,10 @@ static uint64_t tick_sources(uint64_t cur_time, uint64_t last_time)
 	/* call the tick function of each source */
 	source = data->first_source;
 	while (source) {
-		if (source->refs)
-			obs_source_video_tick(source, seconds);
+		obs_source_video_tick(source, seconds);
 		source = (struct obs_source*)source->context.next;
 	}
 
-	/* calculate source volumes */
-	pthread_mutex_lock(&view->channels_mutex);
-
-	source = data->first_source;
-	while (source) {
-		if (source->refs)
-			calculate_base_volume(data, view, source);
-		source = (struct obs_source*)source->context.next;
-	}
-
-	pthread_mutex_unlock(&view->channels_mutex);
-
 	pthread_mutex_unlock(&data->sources_mutex);
 
 	return cur_time;
@@ -115,9 +72,6 @@ static inline void render_displays(void)
 
 	pthread_mutex_unlock(&obs->data.displays_mutex);
 
-	/* render main display */
-	render_display(&obs->video.main_display);
-
 	gs_leave_context();
 }
 
@@ -138,9 +92,12 @@ static inline void unmap_last_surface(struct obs_core_video *video)
 	}
 }
 
+static const char *render_main_texture_name = "render_main_texture";
 static inline void render_main_texture(struct obs_core_video *video,
 		int cur_texture)
 {
+	profile_start(render_main_texture_name);
+
 	struct vec4 clear_color;
 	vec4_set(&clear_color, 0.0f, 0.0f, 0.0f, 1.0f);
 
@@ -151,11 +108,21 @@ static inline void render_main_texture(struct obs_core_video *video,
 	obs_view_render(&obs->data.main_view);
 
 	video->textures_rendered[cur_texture] = true;
+
+	profile_end(render_main_texture_name);
 }
 
 static inline gs_effect_t *get_scale_effect_internal(
 		struct obs_core_video *video)
 {
+	/* if the dimension is under half the size of the original image,
+	 * bicubic/lanczos can't sample enough pixels to create an accurate
+	 * image, so use the bilinear low resolution effect instead */
+	if (video->output_width  < (video->base_width  / 2) &&
+	    video->output_height < (video->base_height / 2)) {
+		return video->bilinear_lowres_effect;
+	}
+
 	switch (video->scale_type) {
 	case OBS_SCALE_BILINEAR: return video->default_effect;
 	case OBS_SCALE_LANCZOS:  return video->lanczos_effect;
@@ -191,9 +158,12 @@ static inline gs_effect_t *get_scale_effect(struct obs_core_video *video,
 	}
 }
 
+static const char *render_output_texture_name = "render_output_texture";
 static inline void render_output_texture(struct obs_core_video *video,
 		int cur_texture, int prev_texture)
 {
+	profile_start(render_output_texture_name);
+
 	gs_texture_t *texture = video->render_textures[prev_texture];
 	gs_texture_t *target  = video->output_textures[cur_texture];
 	uint32_t     width   = gs_texture_get_width(target);
@@ -214,7 +184,7 @@ static inline void render_output_texture(struct obs_core_video *video,
 	size_t      passes, i;
 
 	if (!video->textures_rendered[prev_texture])
-		return;
+		goto end;
 
 	gs_set_render_target(target, NULL);
 	set_render_size(width, height);
@@ -236,6 +206,9 @@ static inline void render_output_texture(struct obs_core_video *video,
 	gs_enable_blending(true);
 
 	video->textures_output[cur_texture] = true;
+
+end:
+	profile_end(render_output_texture_name);
 }
 
 static inline void set_eparam(gs_effect_t *effect, const char *name, float val)
@@ -244,9 +217,12 @@ static inline void set_eparam(gs_effect_t *effect, const char *name, float val)
 	gs_effect_set_float(param, val);
 }
 
+static const char *render_convert_texture_name = "render_convert_texture";
 static void render_convert_texture(struct obs_core_video *video,
 		int cur_texture, int prev_texture)
 {
+	profile_start(render_convert_texture_name);
+
 	gs_texture_t *texture = video->output_textures[prev_texture];
 	gs_texture_t *target  = video->convert_textures[cur_texture];
 	float        fwidth  = (float)video->output_width;
@@ -259,7 +235,7 @@ static void render_convert_texture(struct obs_core_video *video,
 			video->conversion_tech);
 
 	if (!video->textures_output[prev_texture])
-		return;
+		goto end;
 
 	set_eparam(effect, "u_plane_offset", (float)video->plane_offsets[1]);
 	set_eparam(effect, "v_plane_offset", (float)video->plane_offsets[2]);
@@ -290,11 +266,17 @@ static void render_convert_texture(struct obs_core_video *video,
 	gs_enable_blending(true);
 
 	video->textures_converted[cur_texture] = true;
+
+end:
+	profile_end(render_convert_texture_name);
 }
 
+static const char *stage_output_texture_name = "stage_output_texture";
 static inline void stage_output_texture(struct obs_core_video *video,
 		int cur_texture, int prev_texture)
 {
+	profile_start(stage_output_texture_name);
+
 	gs_texture_t   *texture;
 	bool        texture_ready;
 	gs_stagesurf_t *copy = video->copy_surfaces[cur_texture];
@@ -310,11 +292,14 @@ static inline void stage_output_texture(struct obs_core_video *video,
 	unmap_last_surface(video);
 
 	if (!texture_ready)
-		return;
+		goto end;
 
 	gs_stage_texture(copy, texture);
 
 	video->textures_copied[cur_texture] = true;
+
+end:
+	profile_end(stage_output_texture_name);
 }
 
 static inline void render_video(struct obs_core_video *video, int cur_texture,
@@ -449,11 +434,36 @@ static void convert_frame(
 				0, info->height,
 				output->data, output->linesize);
 
+	} else if (info->format == VIDEO_FORMAT_I444) {
+		convert_uyvx_to_i444(
+				input->data[0], input->linesize[0],
+				0, info->height,
+				output->data, output->linesize);
+
 	} else {
 		blog(LOG_ERROR, "convert_frame: unsupported texture format");
 	}
 }
 
+static inline void copy_rgbx_frame(
+		struct video_frame *output, const struct video_data *input,
+		const struct video_output_info *info)
+{
+	uint8_t *in_ptr = input->data[0];
+	uint8_t *out_ptr = output->data[0];
+
+	/* if the line sizes match, do a single copy */
+	if (input->linesize[0] == output->linesize[0]) {
+		memcpy(out_ptr, in_ptr, input->linesize[0] * info->height);
+	} else {
+		for (size_t y = 0; y < info->height; y++) {
+			memcpy(out_ptr, in_ptr, info->width * 4);
+			in_ptr += input->linesize[0];
+			out_ptr += output->linesize[0];
+		}
+	}
+}
+
 static inline void output_video_data(struct obs_core_video *video,
 		struct video_data *input_frame, int count)
 {
@@ -472,6 +482,8 @@ static inline void output_video_data(struct obs_core_video *video,
 
 		} else if (format_is_yuv(info->format)) {
 			convert_frame(&output_frame, input_frame, info);
+		} else {
+			copy_rgbx_frame(&output_frame, input_frame, info);
 		}
 
 		video_output_unlock_frame(video->video);
@@ -494,13 +506,21 @@ static inline void video_sleep(struct obs_core_video *video,
 		*p_time = cur_time + interval_ns * count;
 	}
 
+	video->total_frames += count;
+	video->lagged_frames += count - 1;
+
 	vframe_info.timestamp = cur_time;
 	vframe_info.count = count;
 	circlebuf_push_back(&video->vframe_info_buffer, &vframe_info,
 			sizeof(vframe_info));
 }
 
-static inline void output_frame(uint64_t *cur_time, uint64_t interval)
+static const char *output_frame_gs_context_name = "gs_context(video->graphics)";
+static const char *output_frame_render_video_name = "render_video";
+static const char *output_frame_download_frame_name = "download_frame";
+static const char *output_frame_gs_flush_name = "gs_flush";
+static const char *output_frame_output_video_data_name = "output_video_data";
+static inline void output_frame(void)
 {
 	struct obs_core_video *video = &obs->video;
 	int cur_texture  = video->cur_texture;
@@ -510,11 +530,23 @@ static inline void output_frame(uint64_t *cur_time, uint64_t interval)
 
 	memset(&frame, 0, sizeof(struct video_data));
 
+	profile_start(output_frame_gs_context_name);
 	gs_enter_context(video->graphics);
+
+	profile_start(output_frame_render_video_name);
 	render_video(video, cur_texture, prev_texture);
+	profile_end(output_frame_render_video_name);
+
+	profile_start(output_frame_download_frame_name);
 	frame_ready = download_frame(video, prev_texture, &frame);
+	profile_end(output_frame_download_frame_name);
+
+	profile_start(output_frame_gs_flush_name);
 	gs_flush();
+	profile_end(output_frame_gs_flush_name);
+
 	gs_leave_context();
+	profile_end(output_frame_gs_context_name);
 
 	if (frame_ready) {
 		struct obs_vframe_info vframe_info;
@@ -522,29 +554,54 @@ static inline void output_frame(uint64_t *cur_time, uint64_t interval)
 				sizeof(vframe_info));
 
 		frame.timestamp = vframe_info.timestamp;
+		profile_start(output_frame_output_video_data_name);
 		output_video_data(video, &frame, vframe_info.count);
+		profile_end(output_frame_output_video_data_name);
 	}
 
 	if (++video->cur_texture == NUM_TEXTURES)
 		video->cur_texture = 0;
-
-	video_sleep(video, cur_time, interval);
 }
 
+#define NBSP "\xC2\xA0"
+
+static const char *tick_sources_name = "tick_sources";
+static const char *render_displays_name = "render_displays";
+static const char *output_frame_name = "output_frame";
 void *obs_video_thread(void *param)
 {
 	uint64_t last_time = 0;
-	uint64_t cur_time = os_gettime_ns();
 	uint64_t interval = video_output_get_frame_time(obs->video.video);
 
+	obs->video.video_time = os_gettime_ns();
+
 	os_set_thread_name("libobs: graphics thread");
 
+	const char *video_thread_name =
+		profile_store_name(obs_get_profiler_name_store(),
+			"obs_video_thread(%g"NBSP"ms)", interval / 1000000.);
+	profile_register_root(video_thread_name, interval);
+
 	while (!video_output_stopped(obs->video.video)) {
-		last_time = tick_sources(cur_time, last_time);
+		profile_start(video_thread_name);
 
+		profile_start(tick_sources_name);
+		last_time = tick_sources(obs->video.video_time, last_time);
+		profile_end(tick_sources_name);
+
+		profile_start(render_displays_name);
 		render_displays();
+		profile_end(render_displays_name);
 
-		output_frame(&cur_time, interval);
+		profile_start(output_frame_name);
+		output_frame();
+		profile_end(output_frame_name);
+
+		profile_end(video_thread_name);
+
+		profile_reenable_thread();
+
+		video_sleep(&obs->video, &obs->video.video_time, interval);
 	}
 
 	UNUSED_PARAMETER(param);
