@@ -28,18 +28,6 @@ struct obs_core *obs = NULL;
 extern void add_default_module_paths(void);
 extern char *find_libobs_data_file(const char *file);
 
-static inline void make_gs_init_data(struct gs_init_data *gid,
-		const struct obs_video_info *ovi)
-{
-	memcpy(&gid->window, &ovi->window, sizeof(struct gs_window));
-	gid->cx              = ovi->window_width;
-	gid->cy              = ovi->window_height;
-	gid->num_backbuffers = 2;
-	gid->format          = GS_RGBA;
-	gid->zsformat        = GS_ZS_NONE;
-	gid->adapter         = ovi->adapter;
-}
-
 static inline void make_video_info(struct video_output_info *vi,
 		struct obs_video_info *ovi)
 {
@@ -118,6 +106,37 @@ static inline void set_nv12_sizes(const struct obs_video_info *ovi)
 	video->conversion_tech = "NV12";
 }
 
+static inline void set_444p_sizes(const struct obs_video_info *ovi)
+{
+	struct obs_core_video *video = &obs->video;
+	uint32_t chroma_pixels;
+	uint32_t total_bytes;
+
+	chroma_pixels = (ovi->output_width * ovi->output_height);
+	chroma_pixels = GET_ALIGN(chroma_pixels, PIXEL_SIZE);
+
+	video->plane_offsets[0] = 0;
+	video->plane_offsets[1] = chroma_pixels;
+	video->plane_offsets[2] = chroma_pixels + chroma_pixels;
+
+	video->plane_linewidth[0] = ovi->output_width;
+	video->plane_linewidth[1] = ovi->output_width;
+	video->plane_linewidth[2] = ovi->output_width;
+
+	video->plane_sizes[0] = chroma_pixels;
+	video->plane_sizes[1] = chroma_pixels;
+	video->plane_sizes[2] = chroma_pixels;
+
+	total_bytes = video->plane_offsets[2] + chroma_pixels;
+
+	video->conversion_height =
+		(total_bytes/PIXEL_SIZE + ovi->output_width-1) /
+		ovi->output_width;
+
+	video->conversion_height = GET_ALIGN(video->conversion_height, 2);
+	video->conversion_tech = "Planar444";
+}
+
 static inline void calc_gpu_conversion_sizes(const struct obs_video_info *ovi)
 {
 	obs->video.conversion_height = 0;
@@ -133,6 +152,9 @@ static inline void calc_gpu_conversion_sizes(const struct obs_video_info *ovi)
 	case VIDEO_FORMAT_NV12:
 		set_nv12_sizes(ovi);
 		break;
+	case VIDEO_FORMAT_I444:
+		set_444p_sizes(ovi);
+		break;
 	}
 }
 
@@ -193,17 +215,27 @@ static bool obs_init_textures(struct obs_video_info *ovi)
 	return true;
 }
 
+gs_effect_t *obs_load_effect(gs_effect_t **effect, const char *file)
+{
+	if (!*effect) {
+		char *filename = find_libobs_data_file(file);
+		*effect = gs_effect_create_from_file(filename, NULL);
+		bfree(filename);
+	}
+
+	return *effect;
+}
+
 static int obs_init_graphics(struct obs_video_info *ovi)
 {
 	struct obs_core_video *video = &obs->video;
-	struct gs_init_data graphics_data;
+	uint8_t transparent_tex_data[2*2*4] = {0};
+	const uint8_t *transparent_tex = transparent_tex_data;
 	bool success = true;
 	int errorcode;
 
-	make_gs_init_data(&graphics_data, ovi);
-
 	errorcode = gs_create(&video->graphics, ovi->graphics_module,
-			&graphics_data);
+			ovi->adapter);
 	if (errorcode != GS_SUCCESS) {
 		switch (errorcode) {
 		case GS_ERROR_MODULE_NOT_FOUND:
@@ -254,6 +286,19 @@ static int obs_init_graphics(struct obs_video_info *ovi)
 			NULL);
 	bfree(filename);
 
+	filename = find_libobs_data_file("bilinear_lowres_scale.effect");
+	video->bilinear_lowres_effect = gs_effect_create_from_file(filename,
+			NULL);
+	bfree(filename);
+
+	filename = find_libobs_data_file("premultiplied_alpha.effect");
+	video->premultiplied_alpha_effect = gs_effect_create_from_file(filename,
+			NULL);
+	bfree(filename);
+
+	obs->video.transparent_texture = gs_texture_create(2, 2, GS_RGBA, 1,
+			&transparent_tex, 0);
+
 	if (!video->default_effect)
 		success = false;
 	if (gs_get_device_type() == GS_DEVICE_OPENGL) {
@@ -266,6 +311,10 @@ static int obs_init_graphics(struct obs_video_info *ovi)
 		success = false;
 	if (!video->conversion_effect)
 		success = false;
+	if (!video->premultiplied_alpha_effect)
+		success = false;
+	if (!video->transparent_texture)
+		success = false;
 
 	gs_leave_context();
 	return success ? OBS_VIDEO_SUCCESS : OBS_VIDEO_FAIL;
@@ -321,12 +370,6 @@ static int obs_init_video(struct obs_video_info *ovi)
 		return OBS_VIDEO_FAIL;
 	}
 
-	if (!obs_display_init(&video->main_display, NULL))
-		return OBS_VIDEO_FAIL;
-
-	video->main_display.cx = ovi->window_width;
-	video->main_display.cy = ovi->window_height;
-
 	gs_enter_context(video->graphics);
 
 	if (ovi->gpu_conversion && !obs_init_gpu_conversion(ovi))
@@ -365,8 +408,6 @@ static void obs_free_video(void)
 	struct obs_core_video *video = &obs->video;
 
 	if (video->video) {
-		obs_display_free(&video->main_display);
-
 		video_output_close(video->video);
 		video->video = NULL;
 
@@ -416,6 +457,8 @@ static void obs_free_graphics(void)
 	if (video->graphics) {
 		gs_enter_context(video->graphics);
 
+		gs_texture_destroy(video->transparent_texture);
+
 		gs_effect_destroy(video->default_effect);
 		gs_effect_destroy(video->default_rect_effect);
 		gs_effect_destroy(video->opaque_effect);
@@ -423,6 +466,7 @@ static void obs_free_graphics(void)
 		gs_effect_destroy(video->conversion_effect);
 		gs_effect_destroy(video->bicubic_effect);
 		gs_effect_destroy(video->lanczos_effect);
+		gs_effect_destroy(video->bilinear_lowres_effect);
 		video->default_effect = NULL;
 
 		gs_leave_context();
@@ -440,7 +484,6 @@ static bool obs_init_audio(struct audio_output_info *ai)
 	/* TODO: sound subsystem */
 
 	audio->user_volume    = 1.0f;
-	audio->present_volume = 1.0f;
 
 	errorcode = audio_output_open(&audio->audio, ai);
 	if (errorcode == AUDIO_OUTPUT_SUCCESS)
@@ -459,6 +502,10 @@ static void obs_free_audio(void)
 	if (audio->audio)
 		audio_output_close(audio->audio);
 
+	circlebuf_free(&audio->buffered_timestamps);
+	da_free(audio->render_order);
+	da_free(audio->root_nodes);
+
 	memset(audio, 0, sizeof(struct obs_core_audio));
 }
 
@@ -475,10 +522,10 @@ static bool obs_init_data(void)
 		return false;
 	if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE) != 0)
 		goto fail;
-	if (pthread_mutex_init(&data->user_sources_mutex, &attr) != 0)
-		goto fail;
 	if (pthread_mutex_init(&data->sources_mutex, &attr) != 0)
 		goto fail;
+	if (pthread_mutex_init(&data->audio_sources_mutex, &attr) != 0)
+		goto fail;
 	if (pthread_mutex_init(&data->displays_mutex, &attr) != 0)
 		goto fail;
 	if (pthread_mutex_init(&data->outputs_mutex, &attr) != 0)
@@ -530,22 +577,14 @@ static void obs_free_data(void)
 
 	blog(LOG_INFO, "Freeing OBS context data");
 
-	if (data->user_sources.num)
-		blog(LOG_INFO, "\t%d user source(s) were remaining",
-				(int)data->user_sources.num);
-
-	while (data->user_sources.num)
-		obs_source_remove(data->user_sources.array[0]);
-	da_free(data->user_sources);
-
 	FREE_OBS_LINKED_LIST(source);
 	FREE_OBS_LINKED_LIST(output);
 	FREE_OBS_LINKED_LIST(encoder);
 	FREE_OBS_LINKED_LIST(display);
 	FREE_OBS_LINKED_LIST(service);
 
-	pthread_mutex_destroy(&data->user_sources_mutex);
 	pthread_mutex_destroy(&data->sources_mutex);
+	pthread_mutex_destroy(&data->audio_sources_mutex);
 	pthread_mutex_destroy(&data->displays_mutex);
 	pthread_mutex_destroy(&data->outputs_mutex);
 	pthread_mutex_destroy(&data->encoders_mutex);
@@ -555,8 +594,9 @@ static void obs_free_data(void)
 static const char *obs_signals[] = {
 	"void source_create(ptr source)",
 	"void source_destroy(ptr source)",
-	"void source_add(ptr source)",
 	"void source_remove(ptr source)",
+	"void source_save(ptr source)",
+	"void source_load(ptr source)",
 	"void source_activate(ptr source)",
 	"void source_deactivate(ptr source)",
 	"void source_show(ptr source)",
@@ -565,10 +605,18 @@ static const char *obs_signals[] = {
 	"void source_volume(ptr source, in out float volume)",
 	"void source_volume_level(ptr source, float level, float magnitude, "
 		"float peak)",
+	"void source_transition_start(ptr source)",
+	"void source_transition_video_stop(ptr source)",
+	"void source_transition_stop(ptr source)",
 
 	"void channel_change(int channel, in out ptr source, ptr prev_source)",
 	"void master_volume(in out float volume)",
 
+	"void hotkey_layout_change()",
+	"void hotkey_register(ptr hotkey)",
+	"void hotkey_unregister(ptr hotkey)",
+	"void hotkey_bindings_changed(ptr hotkey)",
+
 	NULL
 };
 
@@ -585,21 +633,109 @@ static inline bool obs_init_handlers(void)
 	return signal_handler_add_array(obs->signals, obs_signals);
 }
 
+static pthread_once_t obs_pthread_once_init_token = PTHREAD_ONCE_INIT;
+static inline bool obs_init_hotkeys(void)
+{
+	struct obs_core_hotkeys *hotkeys = &obs->hotkeys;
+	pthread_mutexattr_t attr;
+	bool success = false;
+
+	assert(hotkeys != NULL);
+
+	da_init(hotkeys->hotkeys);
+	hotkeys->signals = obs->signals;
+	hotkeys->name_map_init_token = obs_pthread_once_init_token;
+	hotkeys->mute = bstrdup("Mute");
+	hotkeys->unmute = bstrdup("Unmute");
+	hotkeys->push_to_mute = bstrdup("Push-to-mute");
+	hotkeys->push_to_talk = bstrdup("Push-to-talk");
+	hotkeys->sceneitem_show = bstrdup("Show '%1'");
+	hotkeys->sceneitem_hide = bstrdup("Hide '%1'");
+
+	if (!obs_hotkeys_platform_init(hotkeys))
+		return false;
+
+	if (pthread_mutexattr_init(&attr) != 0)
+		return false;
+	if (pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE) != 0)
+		goto fail;
+	if (pthread_mutex_init(&hotkeys->mutex, &attr) != 0)
+		goto fail;
+
+	if (os_event_init(&hotkeys->stop_event, OS_EVENT_TYPE_MANUAL) != 0)
+		goto fail;
+	if (pthread_create(&hotkeys->hotkey_thread, NULL,
+			obs_hotkey_thread, NULL))
+		goto fail;
+
+	hotkeys->hotkey_thread_initialized = true;
+
+	success = true;
+
+fail:
+	pthread_mutexattr_destroy(&attr);
+	return success;
+}
+
+static inline void stop_hotkeys(void)
+{
+	struct obs_core_hotkeys *hotkeys = &obs->hotkeys;
+	void *thread_ret;
+
+	if (hotkeys->hotkey_thread_initialized) {
+		os_event_signal(hotkeys->stop_event);
+		pthread_join(hotkeys->hotkey_thread, &thread_ret);
+		hotkeys->hotkey_thread_initialized = false;
+	}
+
+	os_event_destroy(hotkeys->stop_event);
+	obs_hotkeys_free();
+}
+
+static inline void obs_free_hotkeys(void)
+{
+	struct obs_core_hotkeys *hotkeys = &obs->hotkeys;
+
+	bfree(hotkeys->mute);
+	bfree(hotkeys->unmute);
+	bfree(hotkeys->push_to_mute);
+	bfree(hotkeys->push_to_talk);
+	bfree(hotkeys->sceneitem_show);
+	bfree(hotkeys->sceneitem_hide);
+
+	obs_hotkey_name_map_free();
+
+	obs_hotkeys_platform_free(hotkeys);
+	pthread_mutex_destroy(&hotkeys->mutex);
+}
+
 extern const struct obs_source_info scene_info;
 
 extern void log_system_info(void);
 
-static bool obs_init(const char *locale)
+static bool obs_init(const char *locale, const char *module_config_path,
+		profiler_name_store_t *store)
 {
 	obs = bzalloc(sizeof(struct obs_core));
 
+	obs->name_store_owned = !store;
+	obs->name_store = store ? store : profiler_name_store_create();
+	if (!obs->name_store) {
+		blog(LOG_ERROR, "Couldn't create profiler name store");
+		return false;
+	}
+
 	log_system_info();
 
 	if (!obs_init_data())
 		return false;
 	if (!obs_init_handlers())
 		return false;
+	if (!obs_init_hotkeys())
+		return false;
 
+	if (module_config_path)
+		obs->module_config_path = bstrdup(module_config_path);
 	obs->locale = bstrdup(locale);
 	obs_register_source(&scene_info);
 	add_default_module_paths();
@@ -608,12 +744,18 @@ static bool obs_init(const char *locale)
 
 #ifdef _WIN32
 extern void initialize_crash_handler(void);
+extern void initialize_com(void);
+extern void uninitialize_com(void);
 #endif
 
-bool obs_startup(const char *locale)
+static const char *obs_startup_name = "obs_startup";
+bool obs_startup(const char *locale, const char *module_config_path,
+		profiler_name_store_t *store)
 {
 	bool success;
 
+	profile_start(obs_startup_name);
+
 	if (obs) {
 		blog(LOG_WARNING, "Tried to call obs_startup more than once");
 		return false;
@@ -621,9 +763,11 @@ bool obs_startup(const char *locale)
 
 #ifdef _WIN32
 	initialize_crash_handler();
+	initialize_com();
 #endif
 
-	success = obs_init(locale);
+	success = obs_init(locale, module_config_path, store);
+	profile_end(obs_startup_name);
 	if (!success)
 		obs_shutdown();
 
@@ -637,21 +781,36 @@ void obs_shutdown(void)
 	if (!obs)
 		return;
 
-	da_free(obs->input_types);
-	da_free(obs->filter_types);
-	da_free(obs->encoder_types);
-	da_free(obs->transition_types);
-	da_free(obs->output_types);
-	da_free(obs->service_types);
-	da_free(obs->modal_ui_callbacks);
-	da_free(obs->modeless_ui_callbacks);
+#define FREE_REGISTERED_TYPES(structure, list) \
+	do { \
+		for (size_t i = 0; i < list.num; i++) { \
+			struct structure *item = &list.array[i]; \
+			if (item->type_data && item->free_type_data) \
+				item->free_type_data(item->type_data); \
+		} \
+		da_free(list); \
+	} while (false)
+
+	FREE_REGISTERED_TYPES(obs_source_info, obs->source_types);
+	FREE_REGISTERED_TYPES(obs_source_info, obs->input_types);
+	FREE_REGISTERED_TYPES(obs_source_info, obs->filter_types);
+	FREE_REGISTERED_TYPES(obs_source_info, obs->transition_types);
+	FREE_REGISTERED_TYPES(obs_output_info, obs->output_types);
+	FREE_REGISTERED_TYPES(obs_encoder_info, obs->encoder_types);
+	FREE_REGISTERED_TYPES(obs_service_info, obs->service_types);
+	FREE_REGISTERED_TYPES(obs_modal_ui, obs->modal_ui_callbacks);
+	FREE_REGISTERED_TYPES(obs_modeless_ui, obs->modeless_ui_callbacks);
+
+#undef FREE_REGISTERED_TYPES
 
 	stop_video();
+	stop_hotkeys();
 
+	obs_free_audio();
 	obs_free_data();
 	obs_free_video();
+	obs_free_hotkeys();
 	obs_free_graphics();
-	obs_free_audio();
 	proc_handler_destroy(obs->procs);
 	signal_handler_destroy(obs->signals);
 
@@ -667,9 +826,17 @@ void obs_shutdown(void)
 		free_module_path(obs->module_paths.array+i);
 	da_free(obs->module_paths);
 
+	if (obs->name_store_owned)
+		profiler_name_store_free(obs->name_store);
+
+	bfree(obs->module_config_path);
 	bfree(obs->locale);
 	bfree(obs);
 	obs = NULL;
+
+#ifdef _WIN32
+	uninitialize_com();
+#endif
 }
 
 bool obs_initialized(void)
@@ -749,13 +916,16 @@ int obs_reset_video(struct obs_video_info *ovi)
 		}
 	}
 
+	blog(LOG_INFO, "---------------------------------");
 	blog(LOG_INFO, "video settings reset:\n"
 	               "\tbase resolution:   %dx%d\n"
 	               "\toutput resolution: %dx%d\n"
-	               "\tfps:               %d/%d",
+	               "\tfps:               %d/%d\n"
+	               "\tformat:            %s",
 	               ovi->base_width, ovi->base_height,
 	               ovi->output_width, ovi->output_height,
-	               ovi->fps_num, ovi->fps_den);
+	               ovi->fps_num, ovi->fps_den,
+		       get_video_format_name(ovi->output_format));
 
 	return obs_init_video(ovi);
 }
@@ -778,15 +948,14 @@ bool obs_reset_audio(const struct obs_audio_info *oai)
 	ai.samples_per_sec = oai->samples_per_sec;
 	ai.format = AUDIO_FORMAT_FLOAT_PLANAR;
 	ai.speakers = oai->speakers;
-	ai.buffer_ms = oai->buffer_ms;
+	ai.input_callback = audio_callback;
 
+	blog(LOG_INFO, "---------------------------------");
 	blog(LOG_INFO, "audio settings reset:\n"
 	               "\tsamples per sec: %d\n"
-	               "\tspeakers:        %d\n"
-	               "\tbuffering (ms):  %d\n",
+	               "\tspeakers:        %d",
 	               (int)ai.samples_per_sec,
-	               (int)ai.speakers,
-	               (int)ai.buffer_ms);
+	               (int)ai.speakers);
 
 	return obs_init_audio(&ai);
 }
@@ -831,7 +1000,16 @@ bool obs_get_audio_info(struct obs_audio_info *oai)
 
 	oai->samples_per_sec = info->samples_per_sec;
 	oai->speakers = info->speakers;
-	oai->buffer_ms = info->buffer_ms;
+	return true;
+}
+
+bool obs_enum_source_types(size_t idx, const char **id)
+{
+	if (!obs) return false;
+
+	if (idx >= obs->source_types.num)
+		return false;
+	*id = obs->source_types.array[idx].id;
 	return true;
 }
 
@@ -977,25 +1155,6 @@ void *obs_create_ui(const char *name, const char *task, const char *target,
 	return callback ? callback->create(data, ui_data) : NULL;
 }
 
-bool obs_add_source(obs_source_t *source)
-{
-	struct calldata params = {0};
-
-	if (!obs) return false;
-	if (!source) return false;
-
-	pthread_mutex_lock(&obs->data.sources_mutex);
-	da_push_back(obs->data.user_sources, &source);
-	obs_source_addref(source);
-	pthread_mutex_unlock(&obs->data.sources_mutex);
-
-	calldata_set_ptr(&params, "source", source);
-	signal_handler_signal(obs->signals, "source_add", &params);
-	calldata_free(&params);
-
-	return true;
-}
-
 obs_source_t *obs_get_output_source(uint32_t channel)
 {
 	if (!obs) return NULL;
@@ -1041,17 +1200,25 @@ void obs_set_output_source(uint32_t channel, obs_source_t *source)
 
 void obs_enum_sources(bool (*enum_proc)(void*, obs_source_t*), void *param)
 {
+	obs_source_t *source;
+
 	if (!obs) return;
 
-	pthread_mutex_lock(&obs->data.user_sources_mutex);
+	pthread_mutex_lock(&obs->data.sources_mutex);
+	source = obs->data.first_source;
 
-	for (size_t i = 0; i < obs->data.user_sources.num; i++) {
-		struct obs_source *source = obs->data.user_sources.array[i];
-		if (!enum_proc(param, source))
+	while (source) {
+		obs_source_t *next_source =
+			(obs_source_t*)source->context.next;
+
+		if ((source->info.type == OBS_SOURCE_TYPE_INPUT) != 0 &&
+		    !enum_proc(param, source))
 			break;
+
+		source = next_source;
 	}
 
-	pthread_mutex_unlock(&obs->data.user_sources_mutex);
+	pthread_mutex_unlock(&obs->data.sources_mutex);
 }
 
 static inline void obs_enum(void *pstart, pthread_mutex_t *mutex, void *proc,
@@ -1098,31 +1265,8 @@ void obs_enum_services(bool (*enum_proc)(void*, obs_service_t*), void *param)
 			enum_proc, param);
 }
 
-obs_source_t *obs_get_source_by_name(const char *name)
-{
-	struct obs_core_data *data = &obs->data;
-	struct obs_source *source = NULL;
-	size_t i;
-
-	if (!obs) return NULL;
-
-	pthread_mutex_lock(&data->user_sources_mutex);
-
-	for (i = 0; i < data->user_sources.num; i++) {
-		struct obs_source *cur_source = data->user_sources.array[i];
-		if (strcmp(cur_source->context.name, name) == 0) {
-			source = cur_source;
-			obs_source_addref(source);
-			break;
-		}
-	}
-
-	pthread_mutex_unlock(&data->user_sources_mutex);
-	return source;
-}
-
 static inline void *get_context_by_name(void *vfirst, const char *name,
-		pthread_mutex_t *mutex)
+		pthread_mutex_t *mutex, void *(*addref)(void*))
 {
 	struct obs_context_data **first = vfirst;
 	struct obs_context_data *context;
@@ -1131,8 +1275,10 @@ static inline void *get_context_by_name(void *vfirst, const char *name,
 
 	context = *first;
 	while (context) {
-		if (strcmp(context->name, name) == 0)
+		if (!context->private && strcmp(context->name, name) == 0) {
+			context = addref(context);
 			break;
+		}
 		context = context->next;
 	}
 
@@ -1140,51 +1286,92 @@ static inline void *get_context_by_name(void *vfirst, const char *name,
 	return context;
 }
 
+static inline void *obs_source_addref_safe_(void *ref)
+{
+	return obs_source_get_ref(ref);
+}
+
+static inline void *obs_output_addref_safe_(void *ref)
+{
+	return obs_output_get_ref(ref);
+}
+
+static inline void *obs_encoder_addref_safe_(void *ref)
+{
+	return obs_encoder_get_ref(ref);
+}
+
+static inline void *obs_service_addref_safe_(void *ref)
+{
+	return obs_service_get_ref(ref);
+}
+
+static inline void *obs_id_(void *data)
+{
+	return data;
+}
+
+obs_source_t *obs_get_source_by_name(const char *name)
+{
+	if (!obs) return NULL;
+	return get_context_by_name(&obs->data.first_source, name,
+			&obs->data.sources_mutex, obs_source_addref_safe_);
+}
+
 obs_output_t *obs_get_output_by_name(const char *name)
 {
 	if (!obs) return NULL;
 	return get_context_by_name(&obs->data.first_output, name,
-			&obs->data.outputs_mutex);
+			&obs->data.outputs_mutex, obs_output_addref_safe_);
 }
 
 obs_encoder_t *obs_get_encoder_by_name(const char *name)
 {
 	if (!obs) return NULL;
 	return get_context_by_name(&obs->data.first_encoder, name,
-			&obs->data.encoders_mutex);
+			&obs->data.encoders_mutex, obs_encoder_addref_safe_);
 }
 
 obs_service_t *obs_get_service_by_name(const char *name)
 {
 	if (!obs) return NULL;
 	return get_context_by_name(&obs->data.first_service, name,
-			&obs->data.services_mutex);
+			&obs->data.services_mutex, obs_service_addref_safe_);
 }
 
-gs_effect_t *obs_get_default_effect(void)
+gs_effect_t *obs_get_base_effect(enum obs_base_effect effect)
 {
 	if (!obs) return NULL;
-	return obs->video.default_effect;
+
+	switch (effect) {
+	case OBS_EFFECT_DEFAULT:
+		return obs->video.default_effect;
+	case OBS_EFFECT_DEFAULT_RECT:
+		return obs->video.default_rect_effect;
+	case OBS_EFFECT_OPAQUE:
+		return obs->video.opaque_effect;
+	case OBS_EFFECT_SOLID:
+		return obs->video.solid_effect;
+	case OBS_EFFECT_BICUBIC:
+		return obs->video.bicubic_effect;
+	case OBS_EFFECT_LANCZOS:
+		return obs->video.lanczos_effect;
+	case OBS_EFFECT_BILINEAR_LOWRES:
+		return obs->video.bilinear_lowres_effect;
+	case OBS_EFFECT_PREMULTIPLIED_ALPHA:
+		return obs->video.premultiplied_alpha_effect;
+	}
+
+	return NULL;
 }
 
+/* DEPRECATED */
 gs_effect_t *obs_get_default_rect_effect(void)
 {
 	if (!obs) return NULL;
 	return obs->video.default_rect_effect;
 }
 
-gs_effect_t *obs_get_opaque_effect(void)
-{
-	if (!obs) return NULL;
-	return obs->video.opaque_effect;
-}
-
-gs_effect_t *obs_get_solid_effect(void)
-{
-	if (!obs) return NULL;
-	return obs->video.solid_effect;
-}
-
 signal_handler_t *obs_get_signal_handler(void)
 {
 	if (!obs) return NULL;
@@ -1197,30 +1384,6 @@ proc_handler_t *obs_get_proc_handler(void)
 	return obs->procs;
 }
 
-void obs_add_draw_callback(
-		void (*draw)(void *param, uint32_t cx, uint32_t cy),
-		void *param)
-{
-	if (!obs) return;
-
-	obs_display_add_draw_callback(&obs->video.main_display, draw, param);
-}
-
-void obs_remove_draw_callback(
-		void (*draw)(void *param, uint32_t cx, uint32_t cy),
-		void *param)
-{
-	if (!obs) return;
-
-	obs_display_remove_draw_callback(&obs->video.main_display, draw, param);
-}
-
-void obs_resize(uint32_t cx, uint32_t cy)
-{
-	if (!obs || !obs->video.video || !obs->video.graphics) return;
-	obs_display_resize(&obs->video.main_display, cx, cy);
-}
-
 void obs_render_main_view(void)
 {
 	if (!obs) return;
@@ -1241,36 +1404,29 @@ void obs_set_master_volume(float volume)
 	obs->audio.user_volume = volume;
 }
 
-void obs_set_present_volume(float volume)
-{
-	if (!obs) return;
-	obs->audio.present_volume = volume;
-}
-
 float obs_get_master_volume(void)
 {
 	return obs ? obs->audio.user_volume : 0.0f;
 }
 
-float obs_get_present_volume(void)
-{
-	return obs ? obs->audio.present_volume : 0.0f;
-}
-
-static obs_source_t *obs_load_source_type(obs_data_t *source_data,
-		enum obs_source_type type)
+static obs_source_t *obs_load_source_type(obs_data_t *source_data)
 {
 	obs_data_array_t *filters = obs_data_get_array(source_data, "filters");
 	obs_source_t *source;
 	const char   *name    = obs_data_get_string(source_data, "name");
 	const char   *id      = obs_data_get_string(source_data, "id");
 	obs_data_t   *settings = obs_data_get_obj(source_data, "settings");
+	obs_data_t   *hotkeys  = obs_data_get_obj(source_data, "hotkeys");
 	double       volume;
 	int64_t      sync;
 	uint32_t     flags;
 	uint32_t     mixers;
+	int          di_order;
+	int          di_mode;
 
-	source = obs_source_create(type, id, name, settings);
+	source = obs_source_create(id, name, settings, hotkeys);
+
+	obs_data_release(hotkeys);
 
 	obs_data_set_default_double(source_data, "volume", 1.0);
 	volume = obs_data_get_double(source_data, "volume");
@@ -1294,6 +1450,30 @@ static obs_source_t *obs_load_source_type(obs_data_t *source_data,
 	obs_data_set_default_bool(source_data, "muted", false);
 	obs_source_set_muted(source, obs_data_get_bool(source_data, "muted"));
 
+	obs_data_set_default_bool(source_data, "push-to-mute", false);
+	obs_source_enable_push_to_mute(source,
+			obs_data_get_bool(source_data, "push-to-mute"));
+
+	obs_data_set_default_int(source_data, "push-to-mute-delay", 0);
+	obs_source_set_push_to_mute_delay(source,
+			obs_data_get_int(source_data, "push-to-mute-delay"));
+
+	obs_data_set_default_bool(source_data, "push-to-talk", false);
+	obs_source_enable_push_to_talk(source,
+			obs_data_get_bool(source_data, "push-to-talk"));
+
+	obs_data_set_default_int(source_data, "push-to-talk-delay", 0);
+	obs_source_set_push_to_talk_delay(source,
+			obs_data_get_int(source_data, "push-to-talk-delay"));
+
+	di_mode = (int)obs_data_get_int(source_data, "deinterlace_mode");
+	obs_source_set_deinterlace_mode(source,
+			(enum obs_deinterlace_mode)di_mode);
+
+	di_order = (int)obs_data_get_int(source_data, "deinterlace_field_order");
+	obs_source_set_deinterlace_field_order(source,
+			(enum obs_deinterlace_field_order)di_order);
+
 	if (filters) {
 		size_t count = obs_data_array_count(filters);
 
@@ -1302,7 +1482,7 @@ static obs_source_t *obs_load_source_type(obs_data_t *source_data,
 				obs_data_array_item(filters, i);
 
 			obs_source_t *filter = obs_load_source_type(
-					filter_data, OBS_SOURCE_TYPE_FILTER);
+					filter_data);
 			if (filter) {
 				obs_source_filter_add(source, filter);
 				obs_source_release(filter);
@@ -1321,35 +1501,54 @@ static obs_source_t *obs_load_source_type(obs_data_t *source_data,
 
 obs_source_t *obs_load_source(obs_data_t *source_data)
 {
-	return obs_load_source_type(source_data, OBS_SOURCE_TYPE_INPUT);
+	return obs_load_source_type(source_data);
 }
 
-void obs_load_sources(obs_data_array_t *array)
+void obs_load_sources(obs_data_array_t *array, obs_load_source_cb cb,
+		void *private_data)
 {
+	if (!obs) return;
+
+	struct obs_core_data *data = &obs->data;
+	DARRAY(obs_source_t*) sources;
 	size_t count;
 	size_t i;
 
-	if (!obs) return;
+	da_init(sources);
 
 	count = obs_data_array_count(array);
+	da_reserve(sources, count);
 
-	pthread_mutex_lock(&obs->data.user_sources_mutex);
+	pthread_mutex_lock(&data->sources_mutex);
 
 	for (i = 0; i < count; i++) {
 		obs_data_t   *source_data = obs_data_array_item(array, i);
 		obs_source_t *source      = obs_load_source(source_data);
 
-		obs_add_source(source);
+		da_push_back(sources, &source);
 
-		obs_source_release(source);
 		obs_data_release(source_data);
 	}
 
 	/* tell sources that we want to load */
-	for (i = 0; i < obs->data.user_sources.num; i++)
-		obs_source_load(obs->data.user_sources.array[i]);
+	for (i = 0; i < sources.num; i++) {
+		obs_source_t *source = sources.array[i];
+		obs_data_t *source_data = obs_data_array_item(array, i);
+		if (source) {
+			if (source->info.type == OBS_SOURCE_TYPE_TRANSITION)
+				obs_transition_load(source, source_data);
+			obs_source_load(source);
+			cb(private_data, source);
+		}
+		obs_data_release(source_data);
+	}
 
-	pthread_mutex_unlock(&obs->data.user_sources_mutex);
+	for (i = 0; i < sources.num; i++)
+		obs_source_release(sources.array[i]);
+
+	pthread_mutex_unlock(&data->sources_mutex);
+
+	da_free(sources);
 }
 
 obs_data_t *obs_save_source(obs_source_t *source)
@@ -1357,6 +1556,8 @@ obs_data_t *obs_save_source(obs_source_t *source)
 	obs_data_array_t *filters = obs_data_array_create();
 	obs_data_t *source_data = obs_data_create();
 	obs_data_t *settings    = obs_source_get_settings(source);
+	obs_data_t *hotkey_data = source->context.hotkey_data;
+	obs_data_t *hotkeys;
 	float      volume      = obs_source_get_volume(source);
 	uint32_t   mixers      = obs_source_get_audio_mixers(source);
 	int64_t    sync        = obs_source_get_sync_offset(source);
@@ -1365,8 +1566,22 @@ obs_data_t *obs_save_source(obs_source_t *source)
 	const char *id         = obs_source_get_id(source);
 	bool       enabled     = obs_source_enabled(source);
 	bool       muted       = obs_source_muted(source);
+	bool       push_to_mute= obs_source_push_to_mute_enabled(source);
+	uint64_t   ptm_delay   = obs_source_get_push_to_mute_delay(source);
+	bool       push_to_talk= obs_source_push_to_talk_enabled(source);
+	uint64_t   ptt_delay   = obs_source_get_push_to_talk_delay(source);
+	int        di_mode     = (int)obs_source_get_deinterlace_mode(source);
+	int        di_order    =
+		(int)obs_source_get_deinterlace_field_order(source);
 
 	obs_source_save(source);
+	hotkeys = obs_hotkeys_save_source(source);
+
+	if (hotkeys) {
+		obs_data_release(hotkey_data);
+		source->context.hotkey_data = hotkeys;
+		hotkey_data = hotkeys;
+	}
 
 	obs_data_set_string(source_data, "name",     name);
 	obs_data_set_string(source_data, "id",       id);
@@ -1377,6 +1592,16 @@ obs_data_t *obs_save_source(obs_source_t *source)
 	obs_data_set_double(source_data, "volume",   volume);
 	obs_data_set_bool  (source_data, "enabled",  enabled);
 	obs_data_set_bool  (source_data, "muted",    muted);
+	obs_data_set_bool  (source_data, "push-to-mute", push_to_mute);
+	obs_data_set_int   (source_data, "push-to-mute-delay", ptm_delay);
+	obs_data_set_bool  (source_data, "push-to-talk", push_to_talk);
+	obs_data_set_int   (source_data, "push-to-talk-delay", ptt_delay);
+	obs_data_set_obj   (source_data, "hotkeys",  hotkey_data);
+	obs_data_set_int   (source_data, "deinterlace_mode", di_mode);
+	obs_data_set_int   (source_data, "deinterlace_field_order", di_order);
+
+	if (source->info.type == OBS_SOURCE_TYPE_TRANSITION)
+		obs_transition_save(source, source_data);
 
 	pthread_mutex_lock(&source->filter_mutex);
 
@@ -1399,36 +1624,59 @@ obs_data_t *obs_save_source(obs_source_t *source)
 	return source_data;
 }
 
-obs_data_array_t *obs_save_sources(void)
+obs_data_array_t *obs_save_sources_filtered(obs_save_source_filter_cb cb,
+		void *data_)
 {
-	obs_data_array_t *array;
-	size_t i;
-
 	if (!obs) return NULL;
 
+	struct obs_core_data *data = &obs->data;
+	obs_data_array_t *array;
+	obs_source_t *source;
+
 	array = obs_data_array_create();
 
-	pthread_mutex_lock(&obs->data.user_sources_mutex);
+	pthread_mutex_lock(&data->sources_mutex);
 
-	for (i = 0; i < obs->data.user_sources.num; i++) {
-		obs_source_t *source      = obs->data.user_sources.array[i];
-		obs_data_t   *source_data = obs_save_source(source);
+	source = data->first_source;
 
-		obs_data_array_push_back(array, source_data);
-		obs_data_release(source_data);
+	while (source) {
+		if ((source->info.type != OBS_SOURCE_TYPE_FILTER) != 0 &&
+				!source->context.private && cb(data_, source)) {
+			obs_data_t *source_data = obs_save_source(source);
+
+			obs_data_array_push_back(array, source_data);
+			obs_data_release(source_data);
+		}
+
+		source = (obs_source_t*)source->context.next;
 	}
 
-	pthread_mutex_unlock(&obs->data.user_sources_mutex);
+	pthread_mutex_unlock(&data->sources_mutex);
 
 	return array;
 }
 
-/* ensures that names are never blank */
-static inline char *dup_name(const char *name)
+static bool save_source_filter(void *data, obs_source_t *source)
 {
+	UNUSED_PARAMETER(data);
+	UNUSED_PARAMETER(source);
+	return true;
+}
+
+obs_data_array_t *obs_save_sources(void)
+{
+	return obs_save_sources_filtered(save_source_filter, NULL);
+}
+
+/* ensures that names are never blank */
+static inline char *dup_name(const char *name, bool private)
+{
+	if (private && !name)
+		return NULL;
+
 	if (!name || !*name) {
 		struct dstr unnamed = {0};
-		dstr_printf(&unnamed, "__unnamed%004lld",
+		dstr_printf(&unnamed, "__unnamed%04lld",
 				obs->data.unnamed_index++);
 
 		return unnamed.array;
@@ -1439,11 +1687,16 @@ static inline char *dup_name(const char *name)
 
 static inline bool obs_context_data_init_wrap(
 		struct obs_context_data *context,
+		enum obs_obj_type       type,
 		obs_data_t              *settings,
-		const char              *name)
+		const char              *name,
+		obs_data_t              *hotkey_data,
+		bool                    private)
 {
 	assert(context);
 	memset(context, 0, sizeof(*context));
+	context->private = private;
+	context->type = type;
 
 	pthread_mutex_init_value(&context->rename_cache_mutex);
 	if (pthread_mutex_init(&context->rename_cache_mutex, NULL) < 0)
@@ -1457,17 +1710,22 @@ static inline bool obs_context_data_init_wrap(
 	if (!context->procs)
 		return false;
 
-	context->name     = dup_name(name);
-	context->settings = obs_data_newref(settings);
+	context->name        = dup_name(name, private);
+	context->settings    = obs_data_newref(settings);
+	context->hotkey_data = obs_data_newref(hotkey_data);
 	return true;
 }
 
 bool obs_context_data_init(
 		struct obs_context_data *context,
+		enum obs_obj_type       type,
 		obs_data_t              *settings,
-		const char              *name)
+		const char              *name,
+		obs_data_t              *hotkey_data,
+		bool                    private)
 {
-	if (obs_context_data_init_wrap(context, settings, name)) {
+	if (obs_context_data_init_wrap(context, type, settings, name,
+				hotkey_data, private)) {
 		return true;
 	} else {
 		obs_context_data_free(context);
@@ -1477,6 +1735,7 @@ bool obs_context_data_init(
 
 void obs_context_data_free(struct obs_context_data *context)
 {
+	obs_hotkeys_context_release(context);
 	signal_handler_destroy(context->signals);
 	proc_handler_destroy(context->procs);
 	obs_data_release(context->settings);
@@ -1532,7 +1791,52 @@ void obs_context_data_setname(struct obs_context_data *context,
 
 	if (context->name)
 		da_push_back(context->rename_cache, &context->name);
-	context->name = dup_name(name);
+	context->name = dup_name(name, context->private);
 
 	pthread_mutex_unlock(&context->rename_cache_mutex);
 }
+
+profiler_name_store_t *obs_get_profiler_name_store(void)
+{
+	if (!obs)
+		return NULL;
+
+	return obs->name_store;
+}
+
+uint64_t obs_get_video_frame_time(void)
+{
+	return obs ? obs->video.video_time : 0;
+}
+
+enum obs_obj_type obs_obj_get_type(void *obj)
+{
+	struct obs_context_data *context = obj;
+	return context ? context->type : OBS_OBJ_TYPE_INVALID;
+}
+
+const char *obs_obj_get_id(void *obj)
+{
+	struct obs_context_data *context = obj;
+	if (!context)
+		return NULL;
+
+	switch (context->type) {
+	case OBS_OBJ_TYPE_SOURCE:  return ((obs_source_t*)obj)->info.id;
+	case OBS_OBJ_TYPE_OUTPUT:  return ((obs_output_t*)obj)->info.id;
+	case OBS_OBJ_TYPE_ENCODER: return ((obs_encoder_t*)obj)->info.id;
+	case OBS_OBJ_TYPE_SERVICE: return ((obs_service_t*)obj)->info.id;
+	default:;
+	}
+
+	return NULL;
+}
+
+bool obs_obj_invalid(void *obj)
+{
+	struct obs_context_data *context = obj;
+	if (!context)
+		return true;
+
+	return !context->data;
+}
