@@ -11,7 +11,6 @@
 
 #define HANDLE_RADIUS     4.0f
 #define HANDLE_SEL_RADIUS (HANDLE_RADIUS * 1.5f)
-#define CLAMP_DISTANCE    10.0f
 
 /* TODO: make C++ math classes and clean up code here later */
 
@@ -59,6 +58,7 @@ static bool SceneItemHasVideo(obs_sceneitem_t *item)
 
 static bool CloseFloat(float a, float b, float epsilon=0.01)
 {
+	using std::abs;
 	return abs(a-b) <= epsilon;
 }
 
@@ -132,7 +132,7 @@ static inline vec2 GetOBSScreenSize()
 	return size;
 }
 
-vec3 OBSBasicPreview::GetScreenSnapOffset(const vec3 &tl, const vec3 &br)
+vec3 OBSBasicPreview::GetSnapOffset(const vec3 &tl, const vec3 &br)
 {
 	OBSBasic *main = reinterpret_cast<OBSBasic*>(App()->GetMainWindow());
 	vec2 screenSize = GetOBSScreenSize();
@@ -140,19 +140,50 @@ vec3 OBSBasicPreview::GetScreenSnapOffset(const vec3 &tl, const vec3 &br)
 
 	vec3_zero(&clampOffset);
 
-	const float clampDist = CLAMP_DISTANCE / main->previewScale;
+	const bool snap = config_get_bool(GetGlobalConfig(),
+			"BasicWindow", "SnappingEnabled");
+	if (snap == false)
+		return clampOffset;
 
-	if (fabsf(tl.x) < clampDist)
+	const bool screenSnap = config_get_bool(GetGlobalConfig(),
+			"BasicWindow", "ScreenSnapping");
+	const bool centerSnap = config_get_bool(GetGlobalConfig(),
+			"BasicWindow", "CenterSnapping");
+
+	const float clampDist = config_get_double(GetGlobalConfig(),
+			"BasicWindow", "SnapDistance") / main->previewScale;
+	const float centerX = br.x - (br.x - tl.x) / 2.0f;
+	const float centerY = br.y - (br.y - tl.y) / 2.0f;
+
+	// Left screen edge.
+	if (screenSnap &&
+	    fabsf(tl.x) < clampDist)
 		clampOffset.x = -tl.x;
-	if (fabsf(clampOffset.x) < EPSILON &&
+	// Right screen edge.
+	if (screenSnap &&
+	    fabsf(clampOffset.x) < EPSILON &&
 	    fabsf(screenSize.x - br.x) < clampDist)
 		clampOffset.x = screenSize.x - br.x;
+	// Horizontal center.
+	if (centerSnap &&
+	    fabsf(screenSize.x - (br.x - tl.x)) > clampDist &&
+	    fabsf(screenSize.x / 2.0f - centerX) < clampDist)
+		clampOffset.x = screenSize.x / 2.0f - centerX;
 
-	if (fabsf(tl.y) < clampDist)
+	// Top screen edge.
+	if (screenSnap &&
+	    fabsf(tl.y) < clampDist)
 		clampOffset.y = -tl.y;
-	if (fabsf(clampOffset.y) < EPSILON &&
+	// Bottom screen edge.
+	if (screenSnap &&
+	    fabsf(clampOffset.y) < EPSILON &&
 	    fabsf(screenSize.y - br.y) < clampDist)
 		clampOffset.y = screenSize.y - br.y;
+	// Vertical center.
+	if (centerSnap &&
+	    fabsf(screenSize.y - (br.y - tl.y)) > clampDist &&
+	    fabsf(screenSize.y / 2.0f - centerY) < clampDist)
+		clampOffset.y = screenSize.y / 2.0f - centerY;
 
 	return clampOffset;
 }
@@ -283,11 +314,15 @@ static vec2 GetItemSize(obs_sceneitem_t *item)
 		obs_sceneitem_get_bounds(item, &size);
 	} else {
 		obs_source_t *source = obs_sceneitem_get_source(item);
+		obs_sceneitem_crop crop;
 		vec2 scale;
 
 		obs_sceneitem_get_scale(item, &scale);
-		size.x = float(obs_source_get_width(source))  * scale.x;
-		size.y = float(obs_source_get_height(source)) * scale.y;
+		obs_sceneitem_get_crop(item, &crop);
+		size.x = float(obs_source_get_width(source) -
+				crop.left - crop.right) * scale.x;
+		size.y = float(obs_source_get_height(source) -
+				crop.top - crop.bottom) * scale.y;
 	}
 
 	return size;
@@ -330,6 +365,15 @@ void OBSBasicPreview::GetStretchHandleData(const vec2 &pos)
 				-itemUL.x, -itemUL.y, 0.0f);
 		matrix4_rotate_aa4f(&screenToItem, &screenToItem,
 				0.0f, 0.0f, 1.0f, RAD(-itemRot));
+
+		obs_sceneitem_get_crop(stretchItem, &startCrop);
+		obs_sceneitem_get_pos(stretchItem, &startItemPos);
+
+		obs_source_t *source = obs_sceneitem_get_source(stretchItem);
+		cropSize.x = float(obs_source_get_width(source) -
+				startCrop.left - startCrop.right);
+		cropSize.y = float(obs_source_get_height(source) -
+				startCrop.top - startCrop.bottom);
 	}
 }
 
@@ -339,11 +383,20 @@ void OBSBasicPreview::mousePressEvent(QMouseEvent *event)
 	float pixelRatio = main->devicePixelRatio();
 	float x = float(event->x()) - main->previewX / pixelRatio;
 	float y = float(event->y()) - main->previewY / pixelRatio;
+	Qt::KeyboardModifiers modifiers = QGuiApplication::keyboardModifiers();
+	bool altDown = (modifiers & Qt::AltModifier);
 
-	if (event->button() != Qt::LeftButton)
+	OBSQTDisplay::mousePressEvent(event);
+
+	if (event->button() != Qt::LeftButton &&
+	    event->button() != Qt::RightButton)
 		return;
 
-	mouseDown = true;
+	if (event->button() == Qt::LeftButton)
+		mouseDown = true;
+
+	if (altDown)
+		cropping = true;
 
 	vec2_set(&startPos, x, y);
 	GetStretchHandleData(startPos);
@@ -407,6 +460,7 @@ void OBSBasicPreview::mouseReleaseEvent(QMouseEvent *event)
 		stretchItem = nullptr;
 		mouseDown   = false;
 		mouseMoved  = false;
+		cropping    = false;
 	}
 }
 
@@ -448,6 +502,66 @@ static bool AddItemBounds(obs_scene_t *scene, obs_sceneitem_t *item,
 	return true;
 }
 
+struct OffsetData {
+	float clampDist;
+	vec3 tl, br, offset;
+};
+
+static bool GetSourceSnapOffset(obs_scene_t *scene, obs_sceneitem_t *item,
+		void *param)
+{
+	OffsetData *data = reinterpret_cast<OffsetData*>(param);
+
+	if (obs_sceneitem_selected(item))
+		return true;
+
+	matrix4 boxTransform;
+	obs_sceneitem_get_box_transform(item, &boxTransform);
+
+	vec3 t[4] = {
+		GetTransformedPos(0.0f, 0.0f, boxTransform),
+		GetTransformedPos(1.0f, 0.0f, boxTransform),
+		GetTransformedPos(0.0f, 1.0f, boxTransform),
+		GetTransformedPos(1.0f, 1.0f, boxTransform)
+	};
+
+	bool first = true;
+	vec3 tl, br;
+	vec3_zero(&tl);
+	vec3_zero(&br);
+	for (const vec3 &v : t) {
+		if (first) {
+			vec3_copy(&tl, &v);
+			vec3_copy(&br, &v);
+			first = false;
+		} else {
+			vec3_min(&tl, &tl, &v);
+			vec3_max(&br, &br, &v);
+		}
+	}
+
+	// Snap to other source edges
+#define EDGE_SNAP(l, r, x, y) \
+	do { \
+		double dist = fabsf(l.x - data->r.x); \
+		if (dist < data->clampDist && \
+		    fabsf(data->offset.x) < EPSILON && \
+		    data->tl.y < br.y && \
+		    data->br.y > tl.y && \
+		    (fabsf(data->offset.x) > dist || data->offset.x < EPSILON)) \
+			data->offset.x = l.x - data->r.x; \
+	} while (false)
+
+	EDGE_SNAP(tl, br, x, y);
+	EDGE_SNAP(tl, br, y, x);
+	EDGE_SNAP(br, tl, x, y);
+	EDGE_SNAP(br, tl, y, x);
+#undef EDGE_SNAP
+
+	UNUSED_PARAMETER(scene);
+	return true;
+}
+
 void OBSBasicPreview::SnapItemMovement(vec2 &offset)
 {
 	OBSBasic *main = reinterpret_cast<OBSBasic*>(App()->GetMainWindow());
@@ -461,9 +575,39 @@ void OBSBasicPreview::SnapItemMovement(vec2 &offset)
 	data.br.x += offset.x;
 	data.br.y += offset.y;
 
-	vec3 snapOffset = GetScreenSnapOffset(data.tl, data.br);
-	offset.x += snapOffset.x;
-	offset.y += snapOffset.y;
+	vec3 snapOffset = GetSnapOffset(data.tl, data.br);
+
+	const bool snap = config_get_bool(GetGlobalConfig(),
+			"BasicWindow", "SnappingEnabled");
+	const bool sourcesSnap = config_get_bool(GetGlobalConfig(),
+			"BasicWindow", "SourceSnapping");
+	if (snap == false)
+		return;
+	if (sourcesSnap == false) {
+		offset.x += snapOffset.x;
+		offset.y += snapOffset.y;
+		return;
+	}
+
+	const float clampDist = config_get_double(GetGlobalConfig(),
+			"BasicWindow", "SnapDistance") / main->previewScale;
+
+	OffsetData offsetData;
+	offsetData.clampDist = clampDist;
+	offsetData.tl = data.tl;
+	offsetData.br = data.br;
+	vec3_copy(&offsetData.offset, &snapOffset);
+
+	obs_scene_enum_items(scene, GetSourceSnapOffset, &offsetData);
+
+	if (fabsf(offsetData.offset.x) > EPSILON ||
+	    fabsf(offsetData.offset.y) > EPSILON) {
+		offset.x += offsetData.offset.x;
+		offset.y += offsetData.offset.y;
+	} else {
+		offset.x += snapOffset.x;
+		offset.y += snapOffset.y;
+	}
 }
 
 static bool move_items(obs_scene_t *scene, obs_sceneitem_t *item, void *param)
@@ -582,7 +726,7 @@ void OBSBasicPreview::SnapStretchingToScreen(vec3 &tl, vec3 &br)
 	vec3_max(&boundingBR, &boundingBR, &newBL);
 	vec3_max(&boundingBR, &boundingBR, &newBR);
 
-	vec3 offset = GetScreenSnapOffset(boundingTL, boundingBR);
+	vec3 offset = GetSnapOffset(boundingTL, boundingBR);
 	vec3_add(&offset, &offset, &newTL);
 	vec3_transform(&offset, &offset, &screenToItem);
 	vec3_sub(&offset, &offset, &tl);
@@ -598,6 +742,134 @@ void OBSBasicPreview::SnapStretchingToScreen(vec3 &tl, vec3 &br)
 		br.y += offset.y;
 }
 
+static float maxfunc(float x, float y)
+{
+	return x > y ? x : y;
+}
+
+static float minfunc(float x, float y)
+{
+	return x < y ? x : y;
+}
+
+void OBSBasicPreview::CropItem(const vec2 &pos)
+{
+	obs_bounds_type boundsType = obs_sceneitem_get_bounds_type(stretchItem);
+	uint32_t stretchFlags = (uint32_t)stretchHandle;
+	uint32_t align = obs_sceneitem_get_alignment(stretchItem);
+	vec3 tl, br, pos3;
+
+	if (boundsType != OBS_BOUNDS_NONE) /* TODO */
+		return;
+
+	vec3_zero(&tl);
+	vec3_set(&br, stretchItemSize.x, stretchItemSize.y, 0.0f);
+
+	vec3_set(&pos3, pos.x, pos.y, 0.0f);
+	vec3_transform(&pos3, &pos3, &screenToItem);
+
+	obs_sceneitem_crop crop = startCrop;
+	vec2 scale;
+
+	obs_sceneitem_get_scale(stretchItem, &scale);
+
+	vec2 max_tl;
+	vec2 max_br;
+
+	vec2_set(&max_tl,
+		float(-crop.left) * scale.x,
+		float(-crop.top) * scale.y);
+	vec2_set(&max_br,
+		stretchItemSize.x + crop.right * scale.x,
+		stretchItemSize.y + crop.bottom * scale.y);
+
+	typedef std::function<float (float, float)> minmax_func_t;
+
+	minmax_func_t min_x = scale.x < 0.0f ? maxfunc : minfunc;
+	minmax_func_t min_y = scale.y < 0.0f ? maxfunc : minfunc;
+	minmax_func_t max_x = scale.x < 0.0f ? minfunc : maxfunc;
+	minmax_func_t max_y = scale.y < 0.0f ? minfunc : maxfunc;
+
+	pos3.x = min_x(pos3.x, max_br.x);
+	pos3.x = max_x(pos3.x, max_tl.x);
+	pos3.y = min_y(pos3.y, max_br.y);
+	pos3.y = max_y(pos3.y, max_tl.y);
+
+	if (stretchFlags & ITEM_LEFT) {
+		float maxX = stretchItemSize.x - (2.0 * scale.x);
+		pos3.x = tl.x = min_x(pos3.x, maxX);
+
+	} else if (stretchFlags & ITEM_RIGHT) {
+		float minX = (2.0 * scale.x);
+		pos3.x = br.x = max_x(pos3.x, minX);
+	}
+
+	if (stretchFlags & ITEM_TOP) {
+		float maxY = stretchItemSize.y - (2.0 * scale.y);
+		pos3.y = tl.y = min_y(pos3.y, maxY);
+
+	} else if (stretchFlags & ITEM_BOTTOM) {
+		float minY = (2.0 * scale.y);
+		pos3.y = br.y = max_y(pos3.y, minY);
+	}
+
+#define ALIGN_X (ITEM_LEFT|ITEM_RIGHT)
+#define ALIGN_Y (ITEM_TOP|ITEM_BOTTOM)
+	vec3 newPos;
+	vec3_zero(&newPos);
+
+	uint32_t align_x = (align & ALIGN_X);
+	uint32_t align_y = (align & ALIGN_Y);
+	if (align_x == (stretchFlags & ALIGN_X) && align_x != 0)
+		newPos.x = pos3.x;
+	else if (align & ITEM_RIGHT)
+		newPos.x = stretchItemSize.x;
+	else if (!(align & ITEM_LEFT))
+		newPos.x = stretchItemSize.x * 0.5f;
+
+	if (align_y == (stretchFlags & ALIGN_Y) && align_y != 0)
+		newPos.y = pos3.y;
+	else if (align & ITEM_BOTTOM)
+		newPos.y = stretchItemSize.y;
+	else if (!(align & ITEM_TOP))
+		newPos.y = stretchItemSize.y * 0.5f;
+#undef ALIGN_X
+#undef ALIGN_Y
+
+	crop = startCrop;
+
+	if (stretchFlags & ITEM_LEFT)
+		crop.left += int(std::round(tl.x / scale.x));
+	else if (stretchFlags & ITEM_RIGHT)
+		crop.right += int(std::round((stretchItemSize.x - br.x) / scale.x));
+
+	if (stretchFlags & ITEM_TOP)
+		crop.top += int(std::round(tl.y / scale.y));
+	else if (stretchFlags & ITEM_BOTTOM)
+		crop.bottom += int(std::round((stretchItemSize.y - br.y) / scale.y));
+
+	vec3_transform(&newPos, &newPos, &itemToScreen);
+	newPos.x = std::round(newPos.x);
+	newPos.y = std::round(newPos.y);
+
+#if 0
+	vec3 curPos;
+	vec3_zero(&curPos);
+	obs_sceneitem_get_pos(stretchItem, (vec2*)&curPos);
+	blog(LOG_DEBUG, "curPos {%d, %d} - newPos {%d, %d}",
+			int(curPos.x), int(curPos.y),
+			int(newPos.x), int(newPos.y));
+	blog(LOG_DEBUG, "crop {%d, %d, %d, %d}",
+			crop.left, crop.top,
+			crop.right, crop.bottom);
+#endif
+
+	obs_sceneitem_defer_update_begin(stretchItem);
+	obs_sceneitem_set_crop(stretchItem, &crop);
+	obs_sceneitem_set_pos(stretchItem, (vec2*)&newPos);
+	obs_sceneitem_defer_update_end(stretchItem);
+}
+
 void OBSBasicPreview::StretchItem(const vec2 &pos)
 {
 	Qt::KeyboardModifiers modifiers = QGuiApplication::keyboardModifiers();
@@ -646,6 +918,12 @@ void OBSBasicPreview::StretchItem(const vec2 &pos)
 
 		obs_sceneitem_set_bounds(stretchItem, &size);
 	} else {
+		obs_sceneitem_crop crop;
+		obs_sceneitem_get_crop(stretchItem, &crop);
+
+		baseSize.x -= float(crop.left + crop.right);
+		baseSize.y -= float(crop.top + crop.bottom);
+
 		if (!shiftDown)
 			ClampAspect(tl, br, size, baseSize);
 
@@ -675,10 +953,15 @@ void OBSBasicPreview::mouseMoveEvent(QMouseEvent *event)
 		pos.x = std::round(pos.x);
 		pos.y = std::round(pos.y);
 
-		if (stretchHandle != ItemHandle::None)
-			StretchItem(pos);
-		else if (mouseOverItems)
+		if (stretchHandle != ItemHandle::None) {
+			if (cropping)
+				CropItem(pos);
+			else
+				StretchItem(pos);
+
+		} else if (mouseOverItems) {
 			MoveItems(pos);
+		}
 
 		mouseMoved = true;
 	}
@@ -699,6 +982,14 @@ static void DrawCircleAtPos(float x, float y, matrix4 &matrix,
 	gs_matrix_pop();
 }
 
+static inline bool crop_enabled(const obs_sceneitem_crop *crop)
+{
+	return crop->left > 0  ||
+	       crop->top > 0   ||
+	       crop->right > 0 ||
+	       crop->bottom > 0;
+}
+
 bool OBSBasicPreview::DrawSelectedItem(obs_scene_t *scene,
 		obs_sceneitem_t *item, void *param)
 {
@@ -731,6 +1022,9 @@ bool OBSBasicPreview::DrawSelectedItem(obs_scene_t *scene,
 	if (!visible)
 		return true;
 
+	obs_transform_info info;
+	obs_sceneitem_get_info(item, &info);
+
 	gs_load_vertexbuffer(main->circle);
 
 	DrawCircleAtPos(0.0f, 0.0f, boxTransform, main->previewScale);
@@ -742,12 +1036,36 @@ bool OBSBasicPreview::DrawSelectedItem(obs_scene_t *scene,
 	DrawCircleAtPos(0.5f, 1.0f, boxTransform, main->previewScale);
 	DrawCircleAtPos(1.0f, 0.5f, boxTransform, main->previewScale);
 
-	gs_load_vertexbuffer(main->box);
-
 	gs_matrix_push();
 	gs_matrix_scale3f(main->previewScale, main->previewScale, 1.0f);
 	gs_matrix_mul(&boxTransform);
-	gs_draw(GS_LINESTRIP, 0, 0);
+
+	obs_sceneitem_crop crop;
+	obs_sceneitem_get_crop(item, &crop);
+
+	if (info.bounds_type == OBS_BOUNDS_NONE && crop_enabled(&crop)) {
+		vec4 color;
+		gs_effect_t *eff = gs_get_effect();
+		gs_eparam_t *param = gs_effect_get_param_by_name(eff, "color");
+
+#define DRAW_SIDE(side, vb) \
+		if (crop.side > 0) \
+			vec4_set(&color, 0.0f, 1.0f, 0.0f, 1.0f); \
+		else \
+			vec4_set(&color, 1.0f, 0.0f, 0.0f, 1.0f); \
+		gs_effect_set_vec4(param, &color); \
+		gs_load_vertexbuffer(main->vb); \
+		gs_draw(GS_LINESTRIP, 0, 0);
+
+		DRAW_SIDE(left,   boxLeft);
+		DRAW_SIDE(top,    boxTop);
+		DRAW_SIDE(right,  boxRight);
+		DRAW_SIDE(bottom, boxBottom);
+#undef DRAW_SIDE
+	} else {
+		gs_load_vertexbuffer(main->box);
+		gs_draw(GS_LINESTRIP, 0, 0);
+	}
 
 	gs_matrix_pop();
 
@@ -760,7 +1078,7 @@ void OBSBasicPreview::DrawSceneEditing()
 {
 	OBSBasic *main = reinterpret_cast<OBSBasic*>(App()->GetMainWindow());
 
-	gs_effect_t    *solid = obs_get_solid_effect();
+	gs_effect_t    *solid = obs_get_base_effect(OBS_EFFECT_SOLID);
 	gs_technique_t *tech  = gs_effect_get_technique(solid, "Solid");
 
 	vec4 color;
