@@ -12,6 +12,7 @@ class OBSBasicStatusBar : public QStatusBar {
 	Q_OBJECT
 
 private:
+	QLabel *delayInfo;
 	QLabel *droppedFrames;
 	QLabel *sessionTime;
 	QLabel *cpuUsage;
@@ -19,22 +20,33 @@ private:
 
 	obs_output_t *streamOutput = nullptr;
 	obs_output_t *recordOutput = nullptr;
+	bool active = false;
 
 	int retries = 0;
-	int activeRefs = 0;
 	int totalSeconds = 0;
 
+	int reconnectTimeout = 0;
+
+	int delaySecTotal = 0;
+	int delaySecStarting = 0;
+	int delaySecStopping = 0;
+
+	int startSkippedFrameCount = 0;
+	int startTotalFrameCount = 0;
+	int lastSkippedFrameCount = 0;
+
 	int      bitrateUpdateSeconds = 0;
 	uint64_t lastBytesSent = 0;
 	uint64_t lastBytesSentTime = 0;
 
 	QPointer<QTimer> refreshTimer;
 
-	void DecRef();
-	void IncRef();
-
 	obs_output_t *GetOutput();
 
+	void Activate();
+	void Deactivate();
+
+	void UpdateDelayMsg();
 	void UpdateBandwidth();
 	void UpdateSessionTime();
 	void UpdateDroppedFrames();
@@ -43,7 +55,7 @@ private:
 	static void OBSOutputReconnectSuccess(void *data, calldata_t *params);
 
 private slots:
-	void Reconnect();
+	void Reconnect(int seconds);
 	void ReconnectSuccess();
 	void UpdateStatusBar();
 	void UpdateCPUUsage();
@@ -51,6 +63,8 @@ private slots:
 public:
 	OBSBasicStatusBar(QWidget *parent);
 
+	void StreamDelayStarting(int sec);
+	void StreamDelayStopping(int sec);
 	void StreamStarted(obs_output_t *output);
 	void StreamStopped();
 	void RecordingStarted(obs_output_t *output);
