@@ -113,7 +113,9 @@ static bool find_device_id_by_uid(struct coreaudio_data *ca)
 		success = coreaudio_get_device_id(cf_uid, &ca->device_id);
 	}
 
-	CFRelease(cf_uid);
+	if (cf_uid)
+		CFRelease(cf_uid);
+
 	return success;
 }
 
@@ -502,7 +504,9 @@ static bool coreaudio_get_device_name(struct coreaudio_data *ca)
 	bfree(ca->device_name);
 	ca->device_name = bstrdup(name);
 
-	CFRelease(cf_name);
+	if (cf_name)
+		CFRelease(cf_name);
+
 	return true;
 }
 
@@ -643,13 +647,15 @@ static void coreaudio_uninit(struct coreaudio_data *ca)
 
 /* ------------------------------------------------------------------------- */
 
-static const char *coreaudio_input_getname(void)
+static const char *coreaudio_input_getname(void *unused)
 {
+	UNUSED_PARAMETER(unused);
 	return TEXT_AUDIO_INPUT;
 }
 
-static const char *coreaudio_output_getname(void)
+static const char *coreaudio_output_getname(void *unused)
 {
+	UNUSED_PARAMETER(unused);
 	return TEXT_AUDIO_OUTPUT;
 }
 
@@ -777,7 +783,8 @@ static obs_properties_t *coreaudio_output_properties(void *unused)
 struct obs_source_info coreaudio_input_capture_info = {
 	.id             = "coreaudio_input_capture",
 	.type           = OBS_SOURCE_TYPE_INPUT,
-	.output_flags   = OBS_SOURCE_AUDIO,
+	.output_flags   = OBS_SOURCE_AUDIO |
+	                  OBS_SOURCE_DO_NOT_DUPLICATE,
 	.get_name       = coreaudio_input_getname,
 	.create         = coreaudio_create_input_capture,
 	.destroy        = coreaudio_destroy,
@@ -789,7 +796,8 @@ struct obs_source_info coreaudio_input_capture_info = {
 struct obs_source_info coreaudio_output_capture_info = {
 	.id             = "coreaudio_output_capture",
 	.type           = OBS_SOURCE_TYPE_INPUT,
-	.output_flags   = OBS_SOURCE_AUDIO,
+	.output_flags   = OBS_SOURCE_AUDIO |
+	                  OBS_SOURCE_DO_NOT_DUPLICATE,
 	.get_name       = coreaudio_output_getname,
 	.create         = coreaudio_create_output_capture,
 	.destroy        = coreaudio_destroy,
