@@ -32,13 +32,20 @@
 
 struct ffmpeg_cfg {
 	const char         *url;
+	const char         *format_name;
+	const char         *format_mime_type;
+	const char         *muxer_settings;
 	int                video_bitrate;
 	int                audio_bitrate;
 	const char         *video_encoder;
+	int                video_encoder_id;
 	const char         *audio_encoder;
+	int                audio_encoder_id;
 	const char         *video_settings;
 	const char         *audio_settings;
 	enum AVPixelFormat format;
+	enum AVColorRange  color_range;
+	enum AVColorSpace  color_space;
 	int                scale_width;
 	int                scale_height;
 	int                width;
@@ -53,13 +60,14 @@ struct ffmpeg_data {
 	AVFormatContext    *output;
 	struct SwsContext  *swscale;
 
+	int64_t            total_frames;
 	AVPicture          dst_picture;
 	AVFrame            *vframe;
 	int                frame_size;
-	int                total_frames;
 
 	uint64_t           start_timestamp;
 
+	int64_t            total_samples;
 	uint32_t           audio_samplerate;
 	enum audio_format  audio_format;
 	size_t             audio_planes;
@@ -67,7 +75,6 @@ struct ffmpeg_data {
 	struct circlebuf   excess_frames[MAX_AV_PLANES];
 	uint8_t            *samples[MAX_AV_PLANES];
 	AVFrame            *aframe;
-	int                total_samples;
 
 	struct ffmpeg_cfg  config;
 
@@ -146,7 +153,7 @@ static bool open_video_codec(struct ffmpeg_data *data)
 	char **opts = strlist_split(data->config.video_settings, ' ', false);
 	int ret;
 
-	if (data->vcodec->id == AV_CODEC_ID_H264)
+	if (strcmp(data->vcodec->name, "libx264") == 0)
 		av_opt_set(context->priv_data, "preset", "veryfast", 0);
 
 	if (opts) {
@@ -170,6 +177,8 @@ static bool open_video_codec(struct ffmpeg_data *data)
 	data->vframe->format = context->pix_fmt;
 	data->vframe->width  = context->width;
 	data->vframe->height = context->height;
+	data->vframe->colorspace = data->config.color_space;
+	data->vframe->color_range = data->config.color_range;
 
 	ret = avpicture_alloc(&data->dst_picture, context->pix_fmt,
 			context->width, context->height);
@@ -223,10 +232,13 @@ static bool create_video_stream(struct ffmpeg_data *data)
 	context->bit_rate       = data->config.video_bitrate * 1000;
 	context->width          = data->config.scale_width;
 	context->height         = data->config.scale_height;
-	context->time_base.num  = ovi.fps_den;
-	context->time_base.den  = ovi.fps_num;
+	context->time_base      = (AVRational){ ovi.fps_den, ovi.fps_num };
 	context->gop_size       = 120;
 	context->pix_fmt        = closest_format;
+	context->colorspace     = data->config.color_space;
+	context->color_range    = data->config.color_range;
+
+	data->video->time_base = context->time_base;
 
 	if (data->output->oformat->flags & AVFMT_GLOBALHEADER)
 		context->flags |= CODEC_FLAG_GLOBAL_HEADER;
@@ -301,11 +313,16 @@ static bool create_audio_stream(struct ffmpeg_data *data)
 
 	context              = data->audio->codec;
 	context->bit_rate    = data->config.audio_bitrate * 1000;
+	context->time_base   = (AVRational){ 1, aoi.samples_per_sec };
 	context->channels    = get_audio_channels(aoi.speakers);
 	context->sample_rate = aoi.samples_per_sec;
+	context->channel_layout =
+			av_get_default_channel_layout(context->channels);
 	context->sample_fmt  = data->acodec->sample_fmts ?
 		data->acodec->sample_fmts[0] : AV_SAMPLE_FMT_FLTP;
 
+	data->audio->time_base = context->time_base;
+
 	data->audio_samplerate = aoi.samples_per_sec;
 	data->audio_format = convert_ffmpeg_sample_format(context->sample_fmt);
 	data->audio_planes = get_audio_planes(data->audio_format, aoi.speakers);
@@ -347,13 +364,41 @@ static inline bool open_output_file(struct ffmpeg_data *data)
 		}
 	}
 
-	ret = avformat_write_header(data->output, NULL);
+	strncpy(data->output->filename, data->config.url,
+			sizeof(data->output->filename));
+	data->output->filename[sizeof(data->output->filename) - 1] = 0;
+
+	AVDictionary *dict = NULL;
+	if ((ret = av_dict_parse_string(&dict, data->config.muxer_settings,
+				"=", " ", 0))) {
+		blog(LOG_WARNING, "Failed to parse muxer settings: %s\n%s",
+				av_err2str(ret), data->config.muxer_settings);
+
+		av_dict_free(&dict);
+		return false;
+	}
+
+	if (av_dict_count(dict) > 0) {
+		struct dstr str = {0};
+
+		AVDictionaryEntry *entry = NULL;
+		while ((entry = av_dict_get(dict, "", entry,
+						AV_DICT_IGNORE_SUFFIX)))
+			dstr_catf(&str, "\n\t%s=%s", entry->key, entry->value);
+
+		blog(LOG_INFO, "Using muxer settings:%s", str.array);
+		dstr_free(&str);
+	}
+
+	ret = avformat_write_header(data->output, &dict);
 	if (ret < 0) {
 		blog(LOG_WARNING, "Error opening '%s': %s",
 				data->config.url, av_err2str(ret));
 		return false;
 	}
 
+	av_dict_free(&dict);
+
 	return true;
 }
 
@@ -361,6 +406,13 @@ static void close_video(struct ffmpeg_data *data)
 {
 	avcodec_close(data->video->codec);
 	avpicture_free(&data->dst_picture);
+
+	// This format for some reason derefs video frame
+	// too many times
+	if (data->vcodec->id == AV_CODEC_ID_A64_MULTI ||
+	    data->vcodec->id == AV_CODEC_ID_A64_MULTI5)
+		return;
+
 	av_frame_free(&data->vframe);
 }
 
@@ -394,6 +446,42 @@ static void ffmpeg_data_free(struct ffmpeg_data *data)
 	memset(data, 0, sizeof(struct ffmpeg_data));
 }
 
+static inline const char *safe_str(const char *s)
+{
+	if (s == NULL)
+		return "(NULL)";
+	else
+		return s;
+}
+
+static enum AVCodecID get_codec_id(const char *name, int id)
+{
+	AVCodec *codec;
+
+	if (id != 0)
+		return (enum AVCodecID)id;
+
+	if (!name || !*name)
+		return AV_CODEC_ID_NONE;
+
+	codec = avcodec_find_encoder_by_name(name);
+	if (!codec)
+		return AV_CODEC_ID_NONE;
+
+	return codec->id;
+}
+
+static void set_encoder_ids(struct ffmpeg_data *data)
+{
+	data->output->oformat->video_codec = get_codec_id(
+			data->config.video_encoder,
+			data->config.video_encoder_id);
+
+	data->output->oformat->audio_codec = get_codec_id(
+			data->config.audio_encoder,
+			data->config.audio_encoder_id);
+}
+
 static bool ffmpeg_data_init(struct ffmpeg_data *data,
 		struct ffmpeg_cfg *config)
 {
@@ -410,11 +498,31 @@ static bool ffmpeg_data_init(struct ffmpeg_data *data,
 
 	is_rtmp = (astrcmpi_n(config->url, "rtmp://", 7) == 0);
 
-	avformat_alloc_output_context2(&data->output, NULL,
-			is_rtmp ? "flv" : NULL, data->config.url);
+	AVOutputFormat *output_format = av_guess_format(
+			is_rtmp ? "flv" : data->config.format_name,
+			data->config.url,
+			is_rtmp ? NULL : data->config.format_mime_type);
+
+	if (output_format == NULL) {
+		blog(LOG_WARNING, "Couldn't find matching output format with "
+				" parameters: name=%s, url=%s, mime=%s",
+				safe_str(is_rtmp ?
+					"flv" :	data->config.format_name),
+				safe_str(data->config.url),
+				safe_str(is_rtmp ?
+					NULL : data->config.format_mime_type));
+		goto fail;
+	}
+
+	avformat_alloc_output_context2(&data->output, output_format,
+			NULL, NULL);
+
 	if (is_rtmp) {
 		data->output->oformat->video_codec = AV_CODEC_ID_H264;
 		data->output->oformat->audio_codec = AV_CODEC_ID_AAC;
+	} else {
+		if (data->config.format_name)
+			set_encoder_ids(data);
 	}
 
 	if (!data->output) {
@@ -440,8 +548,9 @@ fail:
 
 /* ------------------------------------------------------------------------- */
 
-static const char *ffmpeg_output_getname(void)
+static const char *ffmpeg_output_getname(void *unused)
 {
+	UNUSED_PARAMETER(unused);
 	return obs_module_text("FFmpegOutput");
 }
 
@@ -480,6 +589,7 @@ fail:
 }
 
 static void ffmpeg_output_stop(void *data);
+static void ffmpeg_deactivate(struct ffmpeg_output *output);
 
 static void ffmpeg_output_destroy(void *data)
 {
@@ -526,6 +636,11 @@ static void receive_video(void *param, struct video_data *frame)
 {
 	struct ffmpeg_output *output = param;
 	struct ffmpeg_data   *data   = &output->ff_data;
+
+	// codec doesn't support video or none configured
+	if (!data->video)
+		return;
+
 	AVCodecContext *context = data->video->codec;
 	AVPacket packet = {0};
 	int ret = 0, got_packet;
@@ -671,6 +786,10 @@ static void receive_audio(void *param, struct audio_data *frame)
 	size_t frame_size_bytes;
 	struct audio_data in;
 
+	// codec doesn't support audio or none configured
+	if (!data->audio)
+		return;
+
 	AVCodecContext *context = data->audio->codec;
 
 	if (!data->start_timestamp)
@@ -693,7 +812,7 @@ static void receive_audio(void *param, struct audio_data *frame)
 	}
 }
 
-static bool process_packet(struct ffmpeg_output *output)
+static int process_packet(struct ffmpeg_output *output)
 {
 	AVPacket packet;
 	bool new_packet = false;
@@ -708,7 +827,7 @@ static bool process_packet(struct ffmpeg_output *output)
 	pthread_mutex_unlock(&output->write_mutex);
 
 	if (!new_packet)
-		return true;
+		return 0;
 
 	/*blog(LOG_DEBUG, "size = %d, flags = %lX, stream = %d, "
 			"packets queued: %lu",
@@ -720,10 +839,10 @@ static bool process_packet(struct ffmpeg_output *output)
 		av_free_packet(&packet);
 		blog(LOG_WARNING, "receive_audio: Error writing packet: %s",
 				av_err2str(ret));
-		return false;
+		return ret;
 	}
 
-	return true;
+	return 0;
 }
 
 static void *write_thread(void *data)
@@ -735,11 +854,18 @@ static void *write_thread(void *data)
 		if (os_event_try(output->stop_event) == 0)
 			break;
 
-		if (!process_packet(output)) {
+		int ret = process_packet(output);
+		if (ret != 0) {
+			int code = OBS_OUTPUT_ERROR;
+
 			pthread_detach(output->write_thread);
 			output->write_thread_active = false;
 
-			ffmpeg_output_stop(output);
+			if (ret == -ENOSPC)
+				code = OBS_OUTPUT_NO_SPACE;
+
+			obs_output_signal_stop(output->output, code);
+			ffmpeg_deactivate(output);
 			break;
 		}
 	}
@@ -748,9 +874,19 @@ static void *write_thread(void *data)
 	return NULL;
 }
 
+static inline const char *get_string_or_null(obs_data_t *settings,
+		const char *name)
+{
+	const char *value = obs_data_get_string(settings, name);
+	if (!value || !strlen(value))
+		return NULL;
+	return value;
+}
+
 static bool try_connect(struct ffmpeg_output *output)
 {
 	video_t *video = obs_output_video(output->output);
+	const struct video_output_info *voi = video_output_get_info(video);
 	struct ffmpeg_cfg config;
 	obs_data_t *settings;
 	bool success;
@@ -758,10 +894,18 @@ static bool try_connect(struct ffmpeg_output *output)
 
 	settings = obs_output_get_settings(output->output);
 	config.url = obs_data_get_string(settings, "url");
+	config.format_name = get_string_or_null(settings, "format_name");
+	config.format_mime_type = get_string_or_null(settings,
+			"format_mime_type");
+	config.muxer_settings = obs_data_get_string(settings, "muxer_settings");
 	config.video_bitrate = (int)obs_data_get_int(settings, "video_bitrate");
 	config.audio_bitrate = (int)obs_data_get_int(settings, "audio_bitrate");
-	config.video_encoder = obs_data_get_string(settings, "video_encoder");
-	config.audio_encoder = obs_data_get_string(settings, "audio_encoder");
+	config.video_encoder = get_string_or_null(settings, "video_encoder");
+	config.video_encoder_id = (int)obs_data_get_int(settings,
+			"video_encoder_id");
+	config.audio_encoder = get_string_or_null(settings, "audio_encoder");
+	config.audio_encoder_id = (int)obs_data_get_int(settings,
+			"audio_encoder_id");
 	config.video_settings = obs_data_get_string(settings, "video_settings");
 	config.audio_settings = obs_data_get_string(settings, "audio_settings");
 	config.scale_width = (int)obs_data_get_int(settings, "scale_width");
@@ -771,6 +915,16 @@ static bool try_connect(struct ffmpeg_output *output)
 	config.format = obs_to_ffmpeg_video_format(
 			video_output_get_format(video));
 
+	if (format_is_yuv(voi->format)) {
+		config.color_range = voi->range == VIDEO_RANGE_FULL ?
+			AVCOL_RANGE_JPEG : AVCOL_RANGE_MPEG;
+		config.color_space = voi->colorspace == VIDEO_CS_709 ?
+			AVCOL_SPC_BT709 : AVCOL_SPC_BT470BG;
+	} else {
+		config.color_range = AVCOL_RANGE_UNSPECIFIED;
+		config.color_space = AVCOL_SPC_RGB;
+	}
+
 	if (config.format == AV_PIX_FMT_NONE) {
 		blog(LOG_DEBUG, "invalid pixel format used for FFmpeg output");
 		return false;
@@ -841,26 +995,30 @@ static void ffmpeg_output_stop(void *data)
 
 	if (output->active) {
 		obs_output_end_data_capture(output->output);
-
-		if (output->write_thread_active) {
-			os_event_signal(output->stop_event);
-			os_sem_post(output->write_sem);
-			pthread_join(output->write_thread, NULL);
-			output->write_thread_active = false;
-		}
-
-		pthread_mutex_lock(&output->write_mutex);
-
-		for (size_t i = 0; i < output->packets.num; i++)
-			av_free_packet(output->packets.array+i);
-		da_free(output->packets);
-
-		pthread_mutex_unlock(&output->write_mutex);
-
-		ffmpeg_data_free(&output->ff_data);
+		ffmpeg_deactivate(output);
 	}
 }
 
+static void ffmpeg_deactivate(struct ffmpeg_output *output)
+{
+	if (output->write_thread_active) {
+		os_event_signal(output->stop_event);
+		os_sem_post(output->write_sem);
+		pthread_join(output->write_thread, NULL);
+		output->write_thread_active = false;
+	}
+
+	pthread_mutex_lock(&output->write_mutex);
+
+	for (size_t i = 0; i < output->packets.num; i++)
+		av_free_packet(output->packets.array+i);
+	da_free(output->packets);
+
+	pthread_mutex_unlock(&output->write_mutex);
+
+	ffmpeg_data_free(&output->ff_data);
+}
+
 struct obs_output_info ffmpeg_output = {
 	.id        = "ffmpeg_output",
 	.flags     = OBS_OUTPUT_AUDIO | OBS_OUTPUT_VIDEO,
