@@ -26,6 +26,12 @@
 #include "librtmp/log.h"
 #include "flv-mux.h"
 
+#ifdef _WIN32
+#include <Iphlpapi.h>
+#else
+#include <sys/ioctl.h>
+#endif
+
 #define do_log(level, format, ...) \
 	blog(level, "[rtmp stream: '%s'] " format, \
 			obs_output_get_name(stream->output), ##__VA_ARGS__)
@@ -35,6 +41,7 @@
 #define debug(format, ...) do_log(LOG_DEBUG,   format, ##__VA_ARGS__)
 
 #define OPT_DROP_THRESHOLD "drop_threshold_ms"
+#define OPT_MAX_SHUTDOWN_TIME_SEC "max_shutdown_time_sec"
 
 //#define TEST_FRAMEDROPS
 
@@ -45,17 +52,21 @@ struct rtmp_stream {
 	struct circlebuf packets;
 	bool             sent_headers;
 
-	bool             connecting;
+	volatile bool    connecting;
 	pthread_t        connect_thread;
 
-	bool             active;
+	volatile bool    active;
+	volatile bool    disconnected;
 	pthread_t        send_thread;
 
+	int              max_shutdown_time_sec;
+
 	os_sem_t         *send_sem;
 	os_event_t       *stop_event;
 
 	struct dstr      path, key;
 	struct dstr      username, password;
+	struct dstr      encoder_name;
 
 	/* frame drop variables */
 	int64_t          drop_threshold_usec;
@@ -70,8 +81,9 @@ struct rtmp_stream {
 	RTMP             rtmp;
 };
 
-static const char *rtmp_stream_getname(void)
+static const char *rtmp_stream_getname(void *unused)
 {
+	UNUSED_PARAMETER(unused);
 	return obs_module_text("RTMPStream");
 }
 
@@ -83,23 +95,65 @@ static void log_rtmp(int level, const char *format, va_list args)
 	blogva(LOG_INFO, format, args);
 }
 
+static inline size_t num_buffered_packets(struct rtmp_stream *stream);
+
 static inline void free_packets(struct rtmp_stream *stream)
 {
+	size_t num_packets;
+
+	pthread_mutex_lock(&stream->packets_mutex);
+
+	num_packets = num_buffered_packets(stream);
+	if (num_packets)
+		info("Freeing %d remaining packets", (int)num_packets);
+
 	while (stream->packets.size) {
 		struct encoder_packet packet;
 		circlebuf_pop_front(&stream->packets, &packet, sizeof(packet));
 		obs_free_encoder_packet(&packet);
 	}
+	pthread_mutex_unlock(&stream->packets_mutex);
 }
 
-static void rtmp_stream_stop(void *data);
+static inline bool stopping(struct rtmp_stream *stream)
+{
+	return os_event_try(stream->stop_event) != EAGAIN;
+}
+
+static inline bool connecting(struct rtmp_stream *stream)
+{
+	return os_atomic_load_bool(&stream->connecting);
+}
+
+static inline bool active(struct rtmp_stream *stream)
+{
+	return os_atomic_load_bool(&stream->active);
+}
+
+static inline bool disconnected(struct rtmp_stream *stream)
+{
+	return os_atomic_load_bool(&stream->disconnected);
+}
 
 static void rtmp_stream_destroy(void *data)
 {
 	struct rtmp_stream *stream = data;
 
-	if (stream->active)
-		rtmp_stream_stop(data);
+	if (stopping(stream) && !connecting(stream)) {
+		pthread_join(stream->send_thread, NULL);
+
+	} else if (connecting(stream) || active(stream)) {
+		if (stream->connecting)
+			pthread_join(stream->connect_thread, NULL);
+
+		os_event_signal(stream->stop_event);
+
+		if (active(stream)) {
+			os_sem_post(stream->send_sem);
+			obs_output_end_data_capture(stream->output);
+			pthread_join(stream->send_thread, NULL);
+		}
+	}
 
 	if (stream) {
 		free_packets(stream);
@@ -107,6 +161,7 @@ static void rtmp_stream_destroy(void *data)
 		dstr_free(&stream->key);
 		dstr_free(&stream->username);
 		dstr_free(&stream->password);
+		dstr_free(&stream->encoder_name);
 		os_event_destroy(stream->stop_event);
 		os_sem_destroy(stream->send_sem);
 		pthread_mutex_destroy(&stream->packets_mutex);
@@ -141,23 +196,19 @@ fail:
 static void rtmp_stream_stop(void *data)
 {
 	struct rtmp_stream *stream = data;
-	void *ret;
+
+	if (stopping(stream))
+		return;
+
+	if (connecting(stream))
+		pthread_join(stream->connect_thread, NULL);
 
 	os_event_signal(stream->stop_event);
 
-	if (stream->connecting)
-		pthread_join(stream->connect_thread, &ret);
-
-	if (stream->active) {
-		obs_output_end_data_capture(stream->output);
+	if (active(stream)) {
 		os_sem_post(stream->send_sem);
-		pthread_join(stream->send_thread, &ret);
-		RTMP_Close(&stream->rtmp);
+		obs_output_end_data_capture(stream->output);
 	}
-
-	os_event_reset(stream->stop_event);
-
-	stream->sent_headers = false;
 }
 
 static inline void set_rtmp_str(AVal *val, const char *str)
@@ -190,13 +241,63 @@ static inline bool get_next_packet(struct rtmp_stream *stream,
 	return new_packet;
 }
 
+static bool discard_recv_data(struct rtmp_stream *stream, size_t size)
+{
+	RTMP *rtmp = &stream->rtmp;
+	uint8_t buf[512];
+#ifdef _WIN32
+	int ret;
+#else
+	ssize_t ret;
+#endif
+
+	do {
+		size_t bytes = size > 512 ? 512 : size;
+		size -= bytes;
+
+#ifdef _WIN32
+		ret = recv(rtmp->m_sb.sb_socket, buf, (int)bytes, 0);
+#else
+		ret = recv(rtmp->m_sb.sb_socket, buf, bytes, 0);
+#endif
+
+		if (ret <= 0) {
+#ifdef _WIN32
+			int error = WSAGetLastError();
+#else
+			int error = errno;
+#endif
+			if (ret < 0) {
+				do_log(LOG_ERROR, "recv error: %d (%d bytes)",
+						error, (int)size);
+			}
+			return false;
+		}
+	} while (size > 0);
+
+	return true;
+}
+
 static int send_packet(struct rtmp_stream *stream,
 		struct encoder_packet *packet, bool is_header, size_t idx)
 {
 	uint8_t *data;
 	size_t  size;
+	int     recv_size = 0;
 	int     ret = 0;
 
+#ifdef _WIN32
+	ret = ioctlsocket(stream->rtmp.m_sb.sb_socket, FIONREAD,
+			(u_long*)&recv_size);
+#else
+	ret = ioctl(stream->rtmp.m_sb.sb_socket, FIONREAD, &recv_size);
+#endif
+
+	if (ret >= 0 && recv_size > 0) {
+		if (!discard_recv_data(stream, (size_t)recv_size))
+			return -1;
+	}
+
 	flv_packet_mux(packet, &data, &size, is_header);
 #ifdef TEST_FRAMEDROPS
 	os_sleep_ms(rand() % 40);
@@ -210,81 +311,105 @@ static int send_packet(struct rtmp_stream *stream,
 	return ret;
 }
 
-static inline void send_headers(struct rtmp_stream *stream);
+static inline bool send_headers(struct rtmp_stream *stream);
 
 static bool send_remaining_packets(struct rtmp_stream *stream)
 {
 	struct encoder_packet packet;
+	uint64_t max_ns = (uint64_t)stream->max_shutdown_time_sec * 1000000000;
+	uint64_t begin_time_ns = os_gettime_ns();
 
-	if (!stream->sent_headers)
-		send_headers(stream);
+	if (!stream->sent_headers) {
+		if (!send_headers(stream))
+			return false;
+	}
 
-	while (get_next_packet(stream, &packet))
+	while (get_next_packet(stream, &packet)) {
 		if (send_packet(stream, &packet, false, packet.track_idx) < 0)
 			return false;
 
+		/* Just disconnect if it takes too long to shut down */
+		if ((os_gettime_ns() - begin_time_ns) > max_ns) {
+			info("Took longer than %d second(s) to shut down, "
+			     "automatically stopping connection",
+			     stream->max_shutdown_time_sec);
+			return false;
+		}
+	}
+
 	return true;
 }
 
 static void *send_thread(void *data)
 {
 	struct rtmp_stream *stream = data;
-	bool disconnected = false;
+
+	os_set_thread_name("rtmp-stream: send_thread");
 
 	while (os_sem_wait(stream->send_sem) == 0) {
 		struct encoder_packet packet;
 
-		if (os_event_try(stream->stop_event) != EAGAIN)
+		if (stopping(stream))
 			break;
 		if (!get_next_packet(stream, &packet))
 			continue;
 
-		if (!stream->sent_headers)
-			send_headers(stream);
+		if (!stream->sent_headers) {
+			if (!send_headers(stream)) {
+				os_atomic_set_bool(&stream->disconnected, true);
+				break;
+			}
+		}
 
 		if (send_packet(stream, &packet, false, packet.track_idx) < 0) {
-			disconnected = true;
+			os_atomic_set_bool(&stream->disconnected, true);
 			break;
 		}
 	}
 
-	if (!disconnected && !send_remaining_packets(stream))
-		disconnected = true;
+	if (!disconnected(stream) && !send_remaining_packets(stream))
+		os_atomic_set_bool(&stream->disconnected, true);
 
-	if (disconnected) {
+	if (disconnected(stream)) {
 		info("Disconnected from %s", stream->path.array);
 		free_packets(stream);
 	} else {
 		info("User stopped the stream");
 	}
 
-	if (os_event_try(stream->stop_event) == EAGAIN) {
+	RTMP_Close(&stream->rtmp);
+
+	if (!stopping(stream)) {
 		pthread_detach(stream->send_thread);
 		obs_output_signal_stop(stream->output, OBS_OUTPUT_DISCONNECTED);
 	}
 
-	stream->active = false;
+	os_event_reset(stream->stop_event);
+	os_atomic_set_bool(&stream->active, false);
 	stream->sent_headers = false;
 	return NULL;
 }
 
-static bool send_meta_data(struct rtmp_stream *stream, size_t idx)
+static bool send_meta_data(struct rtmp_stream *stream, size_t idx, bool *next)
 {
 	uint8_t *meta_data;
 	size_t  meta_data_size;
-	bool success = flv_meta_data(stream->output, &meta_data,
+	bool    success = true;
+
+	*next = flv_meta_data(stream->output, &meta_data,
 			&meta_data_size, false, idx);
 
-	if (success) {
-		RTMP_Write(&stream->rtmp, (char*)meta_data,
-				(int)meta_data_size, (int)idx);
+	if (*next) {
+		success = RTMP_Write(&stream->rtmp, (char*)meta_data,
+				(int)meta_data_size, (int)idx) >= 0;
 		bfree(meta_data);
 	}
 
 	return success;
 }
 
-static bool send_audio_header(struct rtmp_stream *stream, size_t idx)
+static bool send_audio_header(struct rtmp_stream *stream, size_t idx,
+		bool *next)
 {
 	obs_output_t  *context  = stream->output;
 	obs_encoder_t *aencoder = obs_output_get_audio_encoder(context, idx);
@@ -295,16 +420,17 @@ static bool send_audio_header(struct rtmp_stream *stream, size_t idx)
 		.timebase_den = 1
 	};
 
-	if (!aencoder)
-		return false;
+	if (!aencoder) {
+		*next = false;
+		return true;
+	}
 
 	obs_encoder_get_extra_data(aencoder, &header, &packet.size);
 	packet.data = bmemdup(header, packet.size);
-	send_packet(stream, &packet, true, idx);
-	return true;
+	return send_packet(stream, &packet, true, idx) >= 0;
 }
 
-static void send_video_header(struct rtmp_stream *stream)
+static bool send_video_header(struct rtmp_stream *stream)
 {
 	obs_output_t  *context  = stream->output;
 	obs_encoder_t *vencoder = obs_output_get_video_encoder(context);
@@ -319,18 +445,27 @@ static void send_video_header(struct rtmp_stream *stream)
 
 	obs_encoder_get_extra_data(vencoder, &header, &size);
 	packet.size = obs_parse_avc_header(&packet.data, header, size);
-	send_packet(stream, &packet, true, 0);
+	return send_packet(stream, &packet, true, 0) >= 0;
 }
 
-static inline void send_headers(struct rtmp_stream *stream)
+static inline bool send_headers(struct rtmp_stream *stream)
 {
 	stream->sent_headers = true;
 	size_t i = 0;
+	bool next = true;
+	bool fail = false;
 
-	send_audio_header(stream, i++);
-	send_video_header(stream);
+	if (!send_audio_header(stream, i++, &next))
+		return false;
+	if (!send_video_header(stream))
+		return false;
 
-	while (send_audio_header(stream, i++));
+	while (next) {
+		if (!send_audio_header(stream, i++, &next))
+			return false;
+	}
+
+	return true;
 }
 
 static inline bool reset_semaphore(struct rtmp_stream *stream)
@@ -364,6 +499,7 @@ static int init_send(struct rtmp_stream *stream)
 {
 	int ret;
 	size_t idx = 0;
+	bool next = true;
 
 #if defined(_WIN32)
 	adjust_sndbuf_size(stream, MIN_SENDBUF_SIZE);
@@ -378,13 +514,76 @@ static int init_send(struct rtmp_stream *stream)
 		return OBS_OUTPUT_ERROR;
 	}
 
-	stream->active = true;
-	while (send_meta_data(stream, idx++));
+	os_atomic_set_bool(&stream->active, true);
+	while (next) {
+		if (!send_meta_data(stream, idx++, &next)) {
+			warn("Disconnected while attempting to connect to "
+			     "server.");
+			return OBS_OUTPUT_DISCONNECTED;
+		}
+	}
 	obs_output_begin_data_capture(stream->output, 0);
 
 	return OBS_OUTPUT_SUCCESS;
 }
 
+#ifdef _WIN32
+static void win32_log_interface_type(struct rtmp_stream *stream)
+{
+	RTMP *rtmp = &stream->rtmp;
+	MIB_IPFORWARDROW route;
+	uint32_t dest_addr, source_addr;
+	char hostname[256];
+	HOSTENT *h;
+
+	if (rtmp->Link.hostname.av_len >= sizeof(hostname) - 1)
+		return;
+
+	strncpy(hostname, rtmp->Link.hostname.av_val, sizeof(hostname));
+	hostname[rtmp->Link.hostname.av_len] = 0;
+
+	h = gethostbyname(hostname);
+	if (!h)
+		return;
+
+	dest_addr = *(uint32_t*)h->h_addr_list[0];
+
+	if (rtmp->m_bindIP.addrLen == 0)
+		source_addr = 0;
+	else if (rtmp->m_bindIP.addr.ss_family == AF_INET)
+		source_addr = (*(struct sockaddr_in*)&rtmp->m_bindIP)
+			.sin_addr.S_un.S_addr;
+	else
+		return;
+
+	if (!GetBestRoute(dest_addr, source_addr, &route)) {
+		MIB_IFROW row;
+		memset(&row, 0, sizeof(row));
+		row.dwIndex = route.dwForwardIfIndex;
+
+		if (!GetIfEntry(&row)) {
+			uint32_t speed =row.dwSpeed / 1000000;
+			char *type;
+			struct dstr other = {0};
+
+			if (row.dwType == IF_TYPE_ETHERNET_CSMACD) {
+				type = "ethernet";
+			} else if (row.dwType == IF_TYPE_IEEE80211) {
+				type = "802.11";
+			} else {
+				dstr_printf(&other, "type %lu", row.dwType);
+				type = other.array;
+			}
+
+			info("Interface: %s (%s, %lu mbps)", row.bDescr, type,
+					speed);
+
+			dstr_free(&other);
+		}
+	}
+}
+#endif
+
 static int try_connect(struct rtmp_stream *stream)
 {
 	if (dstr_is_empty(&stream->path)) {
@@ -394,16 +593,29 @@ static int try_connect(struct rtmp_stream *stream)
 
 	info("Connecting to RTMP URL %s...", stream->path.array);
 
+	memset(&stream->rtmp.Link, 0, sizeof(stream->rtmp.Link));
 	if (!RTMP_SetupURL(&stream->rtmp, stream->path.array))
 		return OBS_OUTPUT_BAD_PATH;
 
 	RTMP_EnableWrite(&stream->rtmp);
 
+	dstr_copy(&stream->encoder_name, "FMLE/3.0 (compatible; obs-studio/");
+
+#ifdef HAVE_OBSCONFIG_H
+	dstr_cat(&stream->encoder_name, OBS_VERSION);
+#else
+	dstr_catf(&stream->encoder_name, "%d.%d.%d",
+			LIBOBS_API_MAJOR_VER,
+			LIBOBS_API_MINOR_VER,
+			LIBOBS_API_PATCH_VER);
+#endif
+
+	dstr_cat(&stream->encoder_name, "; FMSc/1.0)");
+
 	set_rtmp_dstr(&stream->rtmp.Link.pubUser,   &stream->username);
 	set_rtmp_dstr(&stream->rtmp.Link.pubPasswd, &stream->password);
+	set_rtmp_dstr(&stream->rtmp.Link.flashVer,  &stream->encoder_name);
 	stream->rtmp.Link.swfUrl = stream->rtmp.Link.tcUrl;
-	set_rtmp_str(&stream->rtmp.Link.flashVer,
-			"FMLE/3.0 (compatible; FMSc/1.0)");
 
 	RTMP_AddStream(&stream->rtmp, stream->key.array);
 
@@ -423,6 +635,10 @@ static int try_connect(struct rtmp_stream *stream)
 	stream->rtmp.m_bSendChunkSizeInfo = true;
 	stream->rtmp.m_bUseNagle          = true;
 
+#ifdef _WIN32
+	win32_log_interface_type(stream);
+#endif
+
 	if (!RTMP_Connect(&stream->rtmp, NULL))
 		return OBS_OUTPUT_CONNECT_FAILED;
 	if (!RTMP_ConnectStream(&stream->rtmp, 0))
@@ -433,46 +649,77 @@ static int try_connect(struct rtmp_stream *stream)
 	return init_send(stream);
 }
 
-static void *connect_thread(void *data)
+static bool init_connect(struct rtmp_stream *stream)
 {
-	struct rtmp_stream *stream = data;
-	int ret = try_connect(stream);
-
-	if (ret != OBS_OUTPUT_SUCCESS) {
-		obs_output_signal_stop(stream->output, ret);
-		info("Connection to %s failed: %d", stream->path.array, ret);
-	}
-
-	if (os_event_try(stream->stop_event) == EAGAIN)
-		pthread_detach(stream->connect_thread);
-
-	stream->connecting = false;
-	return NULL;
-}
-
-static bool rtmp_stream_start(void *data)
-{
-	struct rtmp_stream *stream = data;
-	obs_service_t *service = obs_output_get_service(stream->output);
+	obs_service_t *service;
 	obs_data_t *settings;
 
-	if (!obs_output_can_begin_data_capture(stream->output, 0))
-		return false;
-	if (!obs_output_initialize_encoders(stream->output, 0))
+	if (stopping(stream))
+		pthread_join(stream->send_thread, NULL);
+
+	free_packets(stream);
+
+	service = obs_output_get_service(stream->output);
+	if (!service)
 		return false;
 
+	os_atomic_set_bool(&stream->disconnected, false);
 	stream->total_bytes_sent = 0;
 	stream->dropped_frames   = 0;
+	stream->min_drop_dts_usec= 0;
+	stream->min_priority     = 0;
 
 	settings = obs_output_get_settings(stream->output);
 	dstr_copy(&stream->path,     obs_service_get_url(service));
 	dstr_copy(&stream->key,      obs_service_get_key(service));
 	dstr_copy(&stream->username, obs_service_get_username(service));
 	dstr_copy(&stream->password, obs_service_get_password(service));
+	dstr_depad(&stream->path);
+	dstr_depad(&stream->key);
 	stream->drop_threshold_usec =
 		(int64_t)obs_data_get_int(settings, OPT_DROP_THRESHOLD) * 1000;
+	stream->max_shutdown_time_sec =
+		(int)obs_data_get_int(settings, OPT_MAX_SHUTDOWN_TIME_SEC);
 	obs_data_release(settings);
+	return true;
+}
 
+static void *connect_thread(void *data)
+{
+	struct rtmp_stream *stream = data;
+	int ret;
+
+	os_set_thread_name("rtmp-stream: connect_thread");
+
+	if (!init_connect(stream)) {
+		obs_output_signal_stop(stream->output, OBS_OUTPUT_BAD_PATH);
+		return NULL;
+	}
+
+	ret = try_connect(stream);
+
+	if (ret != OBS_OUTPUT_SUCCESS) {
+		obs_output_signal_stop(stream->output, ret);
+		info("Connection to %s failed: %d", stream->path.array, ret);
+	}
+
+	if (!stopping(stream))
+		pthread_detach(stream->connect_thread);
+
+	os_atomic_set_bool(&stream->connecting, false);
+	return NULL;
+}
+
+static bool rtmp_stream_start(void *data)
+{
+	struct rtmp_stream *stream = data;
+
+	if (!obs_output_can_begin_data_capture(stream->output, 0))
+		return false;
+	if (!obs_output_initialize_encoders(stream->output, 0))
+		return false;
+
+	os_atomic_set_bool(&stream->connecting, true);
 	return pthread_create(&stream->connect_thread, NULL, connect_thread,
 			stream) == 0;
 }
@@ -508,7 +755,9 @@ static void drop_frames(struct rtmp_stream *stream)
 
 		last_drop_dts_usec = packet.dts_usec;
 
-		if (packet.type == OBS_ENCODER_AUDIO) {
+		/* do not drop audio data or video keyframes */
+		if (packet.type          == OBS_ENCODER_AUDIO ||
+		    packet.drop_priority == OBS_NAL_PRIORITY_HIGHEST) {
 			circlebuf_push_back(&new_buf, &packet, sizeof(packet));
 
 		} else {
@@ -575,7 +824,10 @@ static void rtmp_stream_data(void *data, struct encoder_packet *packet)
 {
 	struct rtmp_stream    *stream = data;
 	struct encoder_packet new_packet;
-	bool                  added_packet;
+	bool                  added_packet = false;
+
+	if (disconnected(stream))
+		return;
 
 	if (packet->type == OBS_ENCODER_VIDEO)
 		obs_parse_avc_packet(&new_packet, packet);
@@ -584,9 +836,11 @@ static void rtmp_stream_data(void *data, struct encoder_packet *packet)
 
 	pthread_mutex_lock(&stream->packets_mutex);
 
-	added_packet = (packet->type == OBS_ENCODER_VIDEO) ?
-		add_video_packet(stream, &new_packet) :
-		add_packet(stream, &new_packet);
+	if (!disconnected(stream)) {
+		added_packet = (packet->type == OBS_ENCODER_VIDEO) ?
+			add_video_packet(stream, &new_packet) :
+			add_packet(stream, &new_packet);
+	}
 
 	pthread_mutex_unlock(&stream->packets_mutex);
 
@@ -599,6 +853,7 @@ static void rtmp_stream_data(void *data, struct encoder_packet *packet)
 static void rtmp_stream_defaults(obs_data_t *defaults)
 {
 	obs_data_set_default_int(defaults, OPT_DROP_THRESHOLD, 600);
+	obs_data_set_default_int(defaults, OPT_MAX_SHUTDOWN_TIME_SEC, 5);
 }
 
 static obs_properties_t *rtmp_stream_properties(void *unused)
@@ -610,6 +865,7 @@ static obs_properties_t *rtmp_stream_properties(void *unused)
 	obs_properties_add_int(props, OPT_DROP_THRESHOLD,
 			obs_module_text("RTMPStream.DropThreshold"),
 			200, 10000, 100);
+
 	return props;
 }
 
